
#  block2: Efficient MPO implementation of quantum chemistry DMRG
#  Copyright (C) 2020-2021 Huanchen Zhai <hczhai@caltech.edu>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#

"""
DDMRG++ for Green's Function.
using pyscf and block2.

Original version:
     Huanchen Zhai, Nov 5, 2020
Revised: support for mpi
     Huanchen Zhai, Tianyu Zhu Mar 29, 2021
"""

from block2 import SU2, SZ, Global, OpNamesSet, Threading, ThreadingTypes
from block2 import init_memory, release_memory, SiteIndex
from block2 import VectorUInt8, PointGroup, FCIDUMP, QCTypes, SeqTypes, OpNames, Random
from block2 import VectorUBond, VectorDouble, NoiseTypes, DecompositionTypes, EquationTypes
from block2 import OrbitalOrdering, VectorUInt16
import time
import numpy as np

# Set spin-adapted or non-spin-adapted here
SpinLabel = SU2
#SpinLabel = SZ

if SpinLabel == SU2:
    from block2.su2 import HamiltonianQC, SimplifiedMPO, Rule, RuleQC, MPOQC
    from block2.su2 import MPSInfo, MPS, MovingEnvironment, DMRG, Linear, IdentityMPO
    from block2.su2 import OpElement, SiteMPO, NoTransposeRule, PDM1MPOQC, Expect
    from block2.su2 import VectorOpElement, LocalMPO
    from block2.su2 import MPICommunicator
else:
    from block2.sz import HamiltonianQC, SimplifiedMPO, Rule, RuleQC, MPOQC
    from block2.sz import MPSInfo, MPS, MovingEnvironment, DMRG, Linear, IdentityMPO
    from block2.sz import OpElement, SiteMPO, NoTransposeRule, PDM1MPOQC, Expect
    from block2.sz import VectorOpElement, LocalMPO
    from block2.sz import MPICommunicator

MPI = MPICommunicator()


def _print(*args, **kwargs):
    if MPI.rank == 0:
        print(*args, **kwargs)


class GFDMRGError(Exception):
    pass


def orbital_reorder(h1e, g2e, method='gaopt'):
    """
    Find an optimal ordering of orbitals for DMRG.
    Ref: J. Chem. Phys. 142, 034102 (2015)

    Args:
        method :
            'gaopt' - genetic algorithm, take several seconds
            'fiedler' - very fast, may be slightly worse than 'gaopt'

    Return a index array "midx":
        reordered_orb_sym = original_orb_sym[midx]
    """
    n_sites = h1e.shape[0]
    hmat = np.zeros((n_sites, n_sites))
    xmat = np.zeros((n_sites, n_sites))
    from pyscf import ao2mo
    if not isinstance(h1e, tuple):
        hmat[:] = np.abs(h1e[:])
        g2e = ao2mo.restore(1, g2e, n_sites)
        for i in range(0, n_sites):
            for j in range(0, n_sites):
                xmat[i, j] = abs(g2e[i, j, j, i])
    else:
        assert SpinLabel == SZ
        assert isinstance(h1e, tuple) and len(h1e) == 2
        assert isinstance(g2e, tuple) and len(g2e) == 3
        hmat[:] = 0.5 * np.abs(h1e[0][:]) + 0.5 * np.abs(h1e[1][:])
        g2eaa = ao2mo.restore(1, g2e[0], n_sites)
        g2ebb = ao2mo.restore(1, g2e[1], n_sites)
        g2eab = ao2mo.restore(1, g2e[2], n_sites)
        for i in range(0, n_sites):
            for j in range(0, n_sites):
                xmat[i, j] = 0.25 * abs(g2eaa[i, j, j, i]) \
                    + 0.25 * abs(g2ebb[i, j, j, i]) \
                    + 0.5 * abs(g2eab[i, j, j, i])
    kmat = VectorDouble((np.array(hmat) * 1E-7 + np.array(xmat)).flatten())
    if method == 'gaopt':
        n_tasks = 32
        opts = dict(
            n_generations=10000, n_configs=n_sites * 2,
            n_elite=8, clone_rate=0.1, mutate_rate=0.1
        )
        midx, mf = None, None
        for _ in range(0, n_tasks):
            idx = OrbitalOrdering.ga_opt(n_sites, kmat, **opts)
            f = OrbitalOrdering.evaluate(n_sites, kmat, idx)
            idx = np.array(idx)
            if mf is None or f < mf:
                midx, mf = idx, f
    elif method == 'fiedler':
        idx = OrbitalOrdering.fiedler(n_sites, kmat)
        midx = np.array(idx)
    else:
        midx = np.array(range(n_sites))
    return midx


class GFDMRG:
    """
    DDMRG++ for Green's Function for molecules.
    """

    def __init__(self, scratch='./nodex', memory=1 * 1E9, omp_threads=8, verbose=2,
                 print_statistics=True, mpi=None, dctr=True):
        """
        Memory is in bytes.
        verbose = 0 (quiet), 2 (per sweep), 3 (per iteration)
        """

        Random.rand_seed(0)
        isize = min(int(memory * 0.1), 200000000)
        init_memory(isize=isize, dsize=int(memory - isize), save_dir=scratch)
        Global.threading = Threading(
            ThreadingTypes.OperatorBatchedGEMM | ThreadingTypes.Global, omp_threads, omp_threads, 1)
        Global.threading.seq_type = SeqTypes.Tasked
        Global.frame.load_buffering = False
        Global.frame.save_buffering = False
        Global.frame.use_main_stack = False
        Global.frame.minimal_disk_usage = True
        self.fcidump = None
        self.hamil = None
        self.verbose = verbose
        self.scratch = scratch
        self.mpo_orig = None
        self.print_statistics = print_statistics
        self.mpi = mpi
        self.dctr = dctr
        self.idx = None # reorder
        self.ridx = None # inv reorder

        if self.verbose >= 2:
            _print(Global.frame)
            _print(Global.threading)

        if mpi is not None:
            if SpinLabel == SU2:
                from block2.su2 import ParallelRuleQC, ParallelRuleNPDMQC, ParallelRuleSiteQC
                from block2.su2 import ParallelRuleSiteQC, ParallelRuleIdentity
            else:
                from block2.sz import ParallelRuleQC, ParallelRuleNPDMQC
                from block2.sz import ParallelRuleSiteQC, ParallelRuleIdentity
            self.prule = ParallelRuleQC(mpi)
            self.pdmrule = ParallelRuleNPDMQC(mpi)
            self.siterule = ParallelRuleSiteQC(mpi)
            self.identrule = ParallelRuleIdentity(mpi)
        else:
            self.prule = None
            self.pdmrule = None
            self.siterule = None
            self.identrule = None

    def init_hamiltonian_fcidump(self, pg, filename, idx=None):
        """Read integrals from FCIDUMP file."""
        assert self.fcidump is None
        self.fcidump = FCIDUMP()
        self.fcidump.read(filename)
        if idx is not None:
            self.fcidump.reorder(VectorUInt16(idx))
            self.idx = idx
            self.ridx = np.argsort(idx)
        self.orb_sym = VectorUInt8(
            map(PointGroup.swap_d2h, self.fcidump.orb_sym))

        vacuum = SpinLabel(0)
        self.target = SpinLabel(self.fcidump.n_elec, self.fcidump.twos,
                                PointGroup.swap_d2h(self.fcidump.isym))
        self.n_sites = self.fcidump.n_sites

        self.hamil = HamiltonianQC(
            vacuum, self.n_sites, self.orb_sym, self.fcidump)
        assert pg in ["d2h", "c1"]

    def init_hamiltonian(self, pg, n_sites, n_elec, twos, isym, orb_sym,
                         e_core, h1e, g2e, tol=1E-13, idx=None,
                         save_fcidump=None):
        """Initialize integrals using h1e, g2e, etc."""
        assert self.fcidump is None
        self.fcidump = FCIDUMP()
        if not isinstance(h1e, tuple):
            mh1e = np.zeros((n_sites * (n_sites + 1) // 2))
            k = 0
            for i in range(0, n_sites):
                for j in range(0, i + 1):
                    assert abs(h1e[i, j] - h1e[j, i]) < tol
                    mh1e[k] = h1e[i, j]
                    k += 1
            mg2e = g2e.flatten()
            mh1e[np.abs(mh1e) < tol] = 0.0
            mg2e[np.abs(mg2e) < tol] = 0.0
            self.fcidump.initialize_su2(
                n_sites, n_elec, twos, isym, e_core, mh1e, mg2e)
        else:
            assert SpinLabel == SZ
            assert isinstance(h1e, tuple) and len(h1e) == 2
            assert isinstance(g2e, tuple) and len(g2e) == 3
            mh1e_a = np.zeros((n_sites * (n_sites + 1) // 2))
            mh1e_b = np.zeros((n_sites * (n_sites + 1) // 2))
            mh1e = (mh1e_a, mh1e_b)
            for xmh1e, xh1e in zip(mh1e, h1e):
                k = 0
                for i in range(0, n_sites):
                    for j in range(0, i + 1):
                        assert abs(xh1e[i, j] - xh1e[j, i]) < tol
                        xmh1e[k] = xh1e[i, j]
                        k += 1
                xmh1e[np.abs(xmh1e) < tol] = 0.0
            mg2e = tuple(xg2e.flatten() for xg2e in g2e)
            for xmg2e in mg2e:
                xmg2e[np.abs(xmg2e) < tol] = 0.0
            self.fcidump.initialize_sz(
                n_sites, n_elec, twos, isym, e_core, mh1e, mg2e)
        self.fcidump.orb_sym = VectorUInt8(orb_sym)
        if idx is not None:
            self.fcidump.reorder(VectorUInt16(idx))
            self.idx = idx
            self.ridx = np.argsort(idx)
        self.orb_sym = VectorUInt8(
            map(PointGroup.swap_d2h, self.fcidump.orb_sym))

        vacuum = SpinLabel(0)
        self.target = SpinLabel(n_elec, twos, PointGroup.swap_d2h(isym))
        self.n_sites = n_sites

        self.hamil = HamiltonianQC(
            vacuum, self.n_sites, self.orb_sym, self.fcidump)

        if save_fcidump is not None:
            if self.mpi is None or self.mpi.rank == 0:
                self.fcidump.orb_sym = VectorUInt8(orb_sym)
                self.fcidump.write(save_fcidump)
            if self.mpi is not None:
                self.mpi.barrier()
        assert pg in ["d2h", "c1"]

    @staticmethod
    def fmt_size(i, suffix='B'):
        if i < 1000:
            return "%d %s" % (i, suffix)
        else:
            a = 1024
            for pf in "KMGTPEZY":
                p = 2
                for k in [10, 100, 1000]:
                    if i < k * a:
                        return "%%.%df %%s%%s" % p % (i / a, pf, suffix)
                    p -= 1
                a *= 1024
        return "??? " + suffix

    def dmrg(self, bond_dims, noises, n_steps=30, conv_tol=1E-7, cutoff=1E-14, occs=None, bias=1.0):
        """Ground-State DMRG."""

        if self.verbose >= 2:
            _print('>>> START GS-DMRG <<<')
        t = time.perf_counter()

        if self.mpi is not None:
            self.mpi.barrier()

        # MultiMPSInfo
        mps_info = MPSInfo(self.n_sites, self.hamil.vacuum,
                           self.target, self.hamil.basis)
        mps_info.tag = 'KET'
        if occs is None:
            if self.verbose >= 2:
                _print("Using FCI INIT MPS")
            mps_info.set_bond_dimension(bond_dims[0])
        else:
            if self.verbose >= 2:
                _print("Using occupation number INIT MPS")
            if self.idx is not None:
                occs = self.fcidump.reorder(VectorDouble(occs), VectorUInt16(self.idx))
            mps_info.set_bond_dimension_using_occ(
                bond_dims[0], VectorDouble(occs), bias=bias)
        mps = MPS(self.n_sites, 0, 2)
        mps.initialize(mps_info)
        mps.random_canonicalize()

        mps.save_mutable()
        mps.deallocate()
        mps_info.save_mutable()
        mps_info.deallocate_mutable()

        # MPO
        tx = time.perf_counter()
        mpo = MPOQC(self.hamil, QCTypes.Conventional)
        mpo = SimplifiedMPO(mpo, RuleQC(), True, True,
                            OpNamesSet((OpNames.R, OpNames.RD)))
        self.mpo_orig = mpo

        if self.mpi is not None:
            if SpinLabel == SU2:
                from block2.su2 import ParallelMPO
            else:
                from block2.sz import ParallelMPO
            mpo = ParallelMPO(mpo, self.prule)

        if self.verbose >= 3:
            _print('MPO time = ', time.perf_counter() - tx)

        if self.print_statistics:
            _print('GS MPO BOND DIMS = ', ''.join(
                ["%6d" % (x.m * x.n) for x in mpo.left_operator_names]))
            max_d = max(bond_dims)
            mps_info2 = MPSInfo(self.n_sites, self.hamil.vacuum,
                                self.target, self.hamil.basis)
            mps_info2.set_bond_dimension(max_d)
            _, mem2, disk = mpo.estimate_storage(mps_info2, 2)
            _print("GS EST MAX MPS BOND DIMS = ", ''.join(
                ["%6d" % x.n_states_total for x in mps_info2.left_dims]))
            _print("GS EST PEAK MEM = ", GFDMRG.fmt_size(
                mem2), " SCRATCH = ", GFDMRG.fmt_size(disk))
            mps_info2.deallocate_mutable()
            mps_info2.deallocate()

        # DMRG
        me = MovingEnvironment(mpo, mps, mps, "DMRG")
        if self.dctr:
            me.delayed_contraction = OpNamesSet.normal_ops()
            me.cached_contraction = True
        tx = time.perf_counter()
        me.init_environments(self.verbose >= 4)
        if self.verbose >= 3:
            _print('DMRG INIT time = ', time.perf_counter() - tx)
        dmrg = DMRG(me, VectorUBond(bond_dims), VectorDouble(noises))
        dmrg.davidson_soft_max_iter = 4000
        dmrg.noise_type = NoiseTypes.ReducedPerturbative
        dmrg.decomp_type = DecompositionTypes.SVD
        dmrg.iprint = max(self.verbose - 1, 0)
        dmrg.cutoff = cutoff
        dmrg.solve(n_steps, mps.center == 0, conv_tol)

        self.gs_energy = dmrg.energies[-1][0]
        self.bond_dim = bond_dims[-1]

        mps.save_data()
        mps_info.save_data(self.scratch + "/GS_MPS_INFO")
        mps_info.deallocate()

        if self.print_statistics:
            dmain, dseco, imain, iseco = Global.frame.peak_used_memory
            _print("GS PEAK MEM USAGE:",
                   "DMEM = ", GFDMRG.fmt_size(dmain + dseco),
                   "(%.0f%%)" % (dmain * 100 / (dmain + dseco)),
                   "IMEM = ", GFDMRG.fmt_size(imain + iseco),
                   "(%.0f%%)" % (imain * 100 / (imain + iseco)))

        if self.verbose >= 1:
            _print("=== GS Energy = %20.15f" % self.gs_energy)

        if self.verbose >= 2:
            _print('>>> COMPLETE GS-DMRG | Time = %.2f <<<' %
                   (time.perf_counter() - t))

    # one-particle density matrix
    # return value:
    #     pdm[0, :, :] -> <AD_{i,alpha} A_{j,alpha}>
    #     pdm[1, :, :] -> < AD_{i,beta}  A_{j,beta}>
    def get_one_pdm(self):
        if self.verbose >= 2:
            _print('>>> START one-pdm <<<')
        t = time.perf_counter()

        if self.mpi is not None:
            self.mpi.barrier()

        mps_info = MPSInfo(0)
        mps_info.load_data(self.scratch + "/GS_MPS_INFO")
        mps = MPS(mps_info)
        mps.load_data()

        mps.info.load_mutable()
        max_bdim = max([x.n_states_total for x in mps.info.left_dims])
        if mps.info.bond_dim < max_bdim:
            mps.info.bond_dim = max_bdim
        max_bdim = max([x.n_states_total for x in mps.info.right_dims])
        if mps.info.bond_dim < max_bdim:
            mps.info.bond_dim = max_bdim

        if self.mpi is not None:
            if SpinLabel == SU2:
                from block2.su2 import ParallelMPO
            else:
                from block2.sz import ParallelMPO

        # 1PDM MPO
        pmpo = PDM1MPOQC(self.hamil)
        pmpo = SimplifiedMPO(pmpo, RuleQC())

        if self.mpi is not None:
            pmpo = ParallelMPO(pmpo, self.pdmrule)

        # 1PDM
        pme = MovingEnvironment(pmpo, mps, mps, "1PDM")
        pme.init_environments(False)
        expect = Expect(pme, mps.info.bond_dim, mps.info.bond_dim)
        expect.iprint = max(self.verbose - 1, 0)
        expect.solve(True, mps.center == 0)
        if SpinLabel == SU2:
            dmr = expect.get_1pdm_spatial(self.n_sites)
            dm = np.array(dmr).copy()
        else:
            dmr = expect.get_1pdm(self.n_sites)
            dm = np.array(dmr).copy()
            dm = dm.reshape((self.n_sites, 2,
                             self.n_sites, 2))
            dm = np.transpose(dm, (0, 2, 1, 3))

        if self.ridx is not None:
            dm[:, :] = dm[self.ridx, :][:, self.ridx]

        mps.save_data()
        mps_info.deallocate()
        dmr.deallocate()
        pmpo.deallocate()

        if self.verbose >= 2:
            _print('>>> COMPLETE one-pdm | Time = %.2f <<<' %
                   (time.perf_counter() - t))

        if SpinLabel == SU2:
            return np.concatenate([dm[None, :, :], dm[None, :, :]], axis=0) / 2
        else:
            return np.concatenate([dm[None, :, :, 0, 0], dm[None, :, :, 1, 1]], axis=0)

    def save_gs_mps(self, save_dir='./gs_mps'):
        import shutil
        import pickle
        import os
        if self.mpi is None or self.mpi.rank == 0:
            pickle.dump(self.gs_energy, open(
                self.scratch + '/GS_ENERGY', 'wb'))
            for k in os.listdir(self.scratch):
                if '.KET.' in k or k == 'GS_MPS_INFO' or k == 'GS_ENERGY':
                    shutil.copy(self.scratch + "/" + k, save_dir + "/" + k)
        if self.mpi is not None:
            self.mpi.barrier()

    def load_gs_mps(self, load_dir='./gs_mps'):
        import shutil
        import pickle
        import os
        if self.mpi is None or self.mpi.rank == 0:
            for k in os.listdir(load_dir):
                shutil.copy(load_dir + "/" + k, self.scratch + "/" + k)
        if self.mpi is not None:
            self.mpi.barrier()
        self.gs_energy = pickle.load(open(self.scratch + '/GS_ENERGY', 'rb'))

    def greens_function(self, bond_dims, noises, gmres_tol, conv_tol, n_steps,
                        cps_bond_dims, cps_noises, cps_conv_tol, cps_n_steps, idxs,
                        eta, freqs, addition, cutoff=1E-14, diag_only=False,
                        n_off_diag_cg=0, alpha=True, occs=None, bias=1.0, mo_coeff=None):
        """Green's function."""
        ops = [None] * len(idxs)
        rkets = [None] * len(idxs)
        rmpos = [None] * len(idxs)

        if self.mpi is not None:
            self.mpi.barrier()

        if self.mpo_orig is None:
            mpo = MPOQC(self.hamil, QCTypes.Conventional)
            mpo = SimplifiedMPO(mpo, RuleQC(), True, True,
                                OpNamesSet((OpNames.R, OpNames.RD)))
            self.mpo_orig = mpo

        mps_info = MPSInfo(0)
        mps_info.load_data(self.scratch + "/GS_MPS_INFO")
        mps = MPS(mps_info)
        mps.load_data()

        if self.mpi is not None:
            if SpinLabel == SU2:
                from block2.su2 import ParallelMPO
            else:
                from block2.sz import ParallelMPO

        if addition:
            mpo = -1.0 * self.mpo_orig
            mpo.const_e += self.gs_energy
        else:
            mpo = 1.0 * self.mpo_orig
            mpo.const_e -= self.gs_energy

        if self.mpi is not None:
            mpo = ParallelMPO(mpo, self.prule)

        if self.print_statistics:
            _print('GF MPO BOND DIMS = ', ''.join(
                ["%6d" % (x.m * x.n) for x in mpo.left_operator_names]))
            max_d = max(bond_dims)
            mps_info2 = MPSInfo(self.n_sites, self.hamil.vacuum,
                                self.target, self.hamil.basis)
            mps_info2.set_bond_dimension(max_d)
            _, mem2, disk = mpo.estimate_storage(mps_info2, 2)
            _print("GF EST MAX MPS BOND DIMS = ", ''.join(
                ["%6d" % x.n_states_total for x in mps_info2.left_dims]))
            _print("GF EST PEAK MEM = ", GFDMRG.fmt_size(
                mem2), " SCRATCH = ", GFDMRG.fmt_size(disk))
            mps_info2.deallocate_mutable()
            mps_info2.deallocate()

        impo = SimplifiedMPO(IdentityMPO(self.hamil),
                             NoTransposeRule(RuleQC()), True, True, OpNamesSet((OpNames.R, OpNames.RD)))

        if self.mpi is not None:
            impo = ParallelMPO(impo, self.identrule)

        def align_mps_center(ket, ref):
            if self.mpi is not None:
                self.mpi.barrier()
            cf = ket.canonical_form
            if ref.center == 0:
                ket.center += 1
                ket.canonical_form = ket.canonical_form[:-1] + 'S'
                while ket.center != 0:
                    ket.move_left(mpo.tf.opf.cg, self.prule)
            else:
                ket.canonical_form = 'K' + ket.canonical_form[1:]
                while ket.center != ket.n_sites - 1:
                    ket.move_right(mpo.tf.opf.cg, self.prule)
                ket.center -= 1
            if self.verbose >= 2:
                _print('CF = %s --> %s' % (cf, ket.canonical_form))

        if mo_coeff is None:
            if self.ridx is not None:
                gidxs = self.ridx[np.array(idxs)]
            else:
                gidxs = idxs
        else:
            if self.idx is not None:
                mo_coeff = mo_coeff[:, self.idx]
            gidxs = list(range(self.n_sites))
            ops = [None] * self.n_sites
            print('idxs = ', idxs, 'gidxs = ', gidxs)

        for ii, idx in enumerate(gidxs):
            if SpinLabel == SZ:
                if addition:
                    ops[ii] = OpElement(OpNames.C, SiteIndex(
                        (idx, ), (0 if alpha else 1, )), SZ(1, 1 if alpha else -1, self.orb_sym[idx]))
                else:
                    ops[ii] = OpElement(OpNames.D, SiteIndex(
                        (idx, ), (0 if alpha else 1, )), SZ(-1, -1 if alpha else 1, self.orb_sym[idx]))
            else:
                if addition:
                    ops[ii] = OpElement(OpNames.C, SiteIndex(
                        (idx, ), ()), SU2(1, 1, self.orb_sym[idx]))
                else:
                    ops[ii] = OpElement(OpNames.D, SiteIndex(
                        (idx, ), ()), SU2(-1, 1, self.orb_sym[idx]))

        for ii, idx in enumerate(idxs):
            if self.mpi is not None:
                self.mpi.barrier()
            if self.verbose >= 2:
                _print('>>> START Compression Site = %4d <<<' % idx)
            t = time.perf_counter()

            rket_info = MPSInfo(self.n_sites, self.hamil.vacuum,
                                self.target + ops[ii].q_label, self.hamil.basis)
            rket_info.tag = 'DKET%d' % idx
            rket_info.set_bond_dimension(mps.info.bond_dim)
            if occs is None:
                if self.verbose >= 2:
                    _print("Using FCI INIT MPS")
                rket_info.set_bond_dimension(mps.info.bond_dim)
            else:
                if self.verbose >= 2:
                    _print("Using occupation number INIT MPS")
                rket_info.set_bond_dimension_using_occ(
                    mps.info.bond_dim, VectorDouble(occs), bias=bias)
            rkets[ii] = MPS(self.n_sites, mps.center, 2)
            rkets[ii].initialize(rket_info)
            rkets[ii].random_canonicalize()

            rkets[ii].save_mutable()
            rkets[ii].deallocate()
            rket_info.save_mutable()
            rket_info.deallocate_mutable()

            if mo_coeff is None:
                # the mpo and gf are in the same basis
                # the mpo is SiteMPO
                rmpos[ii] = SimplifiedMPO(
                    SiteMPO(self.hamil, ops[ii]), NoTransposeRule(RuleQC()), True, True, OpNamesSet((OpNames.R, OpNames.RD)))
            else:
                # the mpo is in mo basis and gf is in ao basis
                # the mpo is sum of SiteMPO (LocalMPO)
                ao_ops = VectorOpElement([None] * self.n_sites)
                for ix in range(self.n_sites):
                    ao_ops[ix] = ops[ix] * mo_coeff[idx, ix]
                rmpos[ii] = SimplifiedMPO(
                    LocalMPO(self.hamil, ao_ops), NoTransposeRule(RuleQC()), True, True, OpNamesSet((OpNames.R, OpNames.RD)))

            if self.mpi is not None:
                rmpos[ii] = ParallelMPO(rmpos[ii], self.siterule)

            if len(cps_noises) == 1 and cps_noises[0] == 0:
                pme = None
            else:
                pme = MovingEnvironment(mpo, rkets[ii], rkets[ii], "PERT")
                pme.init_environments(False)
            rme = MovingEnvironment(rmpos[ii], rkets[ii], mps, "RHS")
            rme.init_environments(False)
            if self.dctr:
                if pme is not None:
                    pme.delayed_contraction = OpNamesSet.normal_ops()
                rme.delayed_contraction = OpNamesSet.normal_ops()

            cps = Linear(pme, rme, VectorUBond(cps_bond_dims),
                         VectorUBond([mps.info.bond_dim]), VectorDouble(cps_noises))
            cps.noise_type = NoiseTypes.ReducedPerturbative
            cps.decomp_type = DecompositionTypes.SVD
            if pme is not None:
                cps.eq_type = EquationTypes.PerturbativeCompression
            cps.iprint = max(self.verbose - 1, 0)
            cps.cutoff = cutoff
            cps.solve(cps_n_steps, mps.center == 0, cps_conv_tol)

            if self.verbose >= 2:
                _print('>>> COMPLETE Compression Site = %4d | Time = %.2f <<<' %
                       (idx, time.perf_counter() - t))

        gf_mat = np.zeros((len(idxs), len(idxs), len(freqs)), dtype=complex)

        for ii, idx in enumerate(idxs):

            if rkets[ii].center != mps.center:
                align_mps_center(rkets[ii], mps)
            lme = MovingEnvironment(mpo, rkets[ii], rkets[ii], "LHS")
            lme.init_environments(False)
            rme = MovingEnvironment(rmpos[ii], rkets[ii], mps, "RHS")
            rme.init_environments(False)
            if self.dctr:
                lme.delayed_contraction = OpNamesSet.normal_ops()
                rme.delayed_contraction = OpNamesSet.normal_ops()

            linear = Linear(lme, rme, VectorUBond(bond_dims),
                            VectorUBond(cps_bond_dims[-1:]), VectorDouble(noises))
            linear.gf_eta = eta
            linear.linear_conv_thrds = VectorDouble([gmres_tol] * n_steps)
            linear.noise_type = NoiseTypes.ReducedPerturbative
            linear.decomp_type = DecompositionTypes.SVD
            # TZ: Not raising error even if CG is not converged
            max_cg_iter = 20000
            linear.linear_soft_max_iter = max_cg_iter
            linear.linear_max_iter = max_cg_iter + 1000
            linear.eq_type = EquationTypes.GreensFunction
            linear.iprint = max(self.verbose - 1, 0)
            linear.cutoff = cutoff

            for iw, w in enumerate(freqs):

                if self.verbose >= 2:
                    _print('>>>   START  GF OMEGA = %10.5f Site = %4d %4d <<<' %
                           (w, idx, idx))
                t = time.perf_counter()

                linear.tme = None
                linear.noises[0] = noises[0]
                linear.linear_soft_max_iter = max_cg_iter
                linear.noises = VectorDouble(noises)
                linear.bra_bond_dims = VectorUBond(bond_dims)
                linear.gf_omega = w
                linear.solve(n_steps, mps.center == 0, conv_tol)
                min_site = np.argmin(np.array(linear.sweep_targets)[:, 1])
                if mps.center == 0:
                    min_site = mps.n_sites - 2 - min_site
                _print("GF.IMAG MIN SITE = %4d" % min_site)
                rgf, igf = linear.targets[-1]
                gf_mat[ii, ii, iw] = rgf + 1j * igf

                dmain, dseco, imain, iseco = Global.frame.peak_used_memory

                if self.verbose >= 1:
                    _print("=== %3s GF (%4d%4d | OMEGA = %10.5f ) = RE %20.15f + IM %20.15f === T = %7.2f" %
                           ("ADD" if addition else "REM", idx, idx, w, rgf, igf, time.perf_counter() - t))

                if self.verbose >= 2:
                    _print('>>> COMPLETE GF OMEGA = %10.5f Site = %4d %4d | Time = %.2f <<<' %
                           (w, idx, idx, time.perf_counter() - t))

                if diag_only:
                    continue

                for jj, idx2 in enumerate(idxs):

                    if jj > ii and rkets[jj].info.target == rkets[ii].info.target:

                        if rkets[jj].center != rkets[ii].center:
                            align_mps_center(rkets[jj], rkets[ii])

                        if self.verbose >= 2:
                            _print('>>>   START  GF OMEGA = %10.5f Site = %4d %4d <<<' % (
                                w, idx2, idx))
                        t = time.perf_counter()

                        tme = MovingEnvironment(
                            impo, rkets[jj], rkets[ii], "GF")
                        tme.init_environments(False)
                        if self.dctr:
                            tme.delayed_contraction = OpNamesSet.normal_ops()
                        linear.noises = VectorDouble(noises[-1:])
                        linear.bra_bond_dims = VectorUBond(bond_dims[-1:])
                        linear.target_bra_bond_dim = cps_bond_dims[-1]
                        linear.target_ket_bond_dim = cps_bond_dims[-1]
                        linear.tme = tme
                        if n_off_diag_cg == 0:
                            linear.solve(1, mps.center != 0, 0)
                            rgf, igf = linear.targets[-1]
                        else:
                            linear.linear_soft_max_iter = n_off_diag_cg if n_off_diag_cg != -1 else max_cg_iter
                            linear.solve(1, mps.center == 0, 0)
                            if mps.center == 0:
                                rgf, igf = np.array(linear.sweep_targets)[::-1][min_site]
                            else:
                                rgf, igf = np.array(linear.sweep_targets)[min_site]
                        gf_mat[jj, ii, iw] = rgf + 1j * igf
                        gf_mat[ii, jj, iw] = rgf + 1j * igf

                        if self.verbose >= 1:
                            _print("=== %3s GF (%4d%4d | OMEGA = %10.5f ) = RE %20.15f + IM %20.15f === T = %7.2f" %
                                   ("ADD" if addition else "REM", idx2, idx, w, rgf, igf, time.perf_counter() - t))

                        if self.verbose >= 2:
                            _print('>>> COMPLETE GF OMEGA = %10.5f Site = %4d %4d | Time = %.2f <<<' %
                                   (w, idx2, idx, time.perf_counter() - t))
        mps.save_data()
        mps_info.deallocate()

        if self.print_statistics:
            _print("GF PEAK MEM USAGE:",
                   "DMEM = ", GFDMRG.fmt_size(dmain + dseco),
                   "(%.0f%%)" % (dmain * 100 / (dmain + dseco)),
                   "IMEM = ", GFDMRG.fmt_size(imain + iseco),
                   "(%.0f%%)" % (imain * 100 / (imain + iseco)))

        return gf_mat

    def __del__(self):
        if self.hamil is not None:
            self.hamil.deallocate()
        if self.fcidump is not None:
            self.fcidump.deallocate()
        if self.mpo_orig is not None:
            self.mpo_orig.deallocate()
        release_memory()


def dmrg_mo_gf(mf, freqs, delta, ao_orbs=None, mo_orbs=None, gmres_tol=1E-7, add_rem='+-',
               n_threads=8, memory=1E9, verbose=1, ignore_ecore=True,
               gs_bond_dims=[500], gs_noises=[1E-5, 1E-5, 1E-6, 1E-7, 0], gs_tol=1E-10, gs_n_steps=30,
               cps_bond_dims=[500], cps_noises=[0], cps_tol=1E-10, cps_n_steps=30,
               gf_bond_dims=[750], gf_noises=[1E-5, 0], gf_tol=1E-8, gf_n_steps=20, scratch='./tmp',
               mo_basis=True, load_dir=None, save_dir=None, pdm_return=True, reorder_method=None,
               lowdin=False, diag_only=False, alpha=True, occs=None, bias=1.0, cutoff=1E-14,
               n_off_diag_cg=-1, mpi=None):
    '''
    Calculate the DMRG GF matrix in the MO basis.

    Args:
        mf : scf object
        freqs : np.ndarray of frequencies (real)
        delta : broadening (real)
        ao_orbs : list of indices of atomic orbtials (if not None, gf will be in ao basis)
        mo_orbs : list of indices of molecular orbtials (if not None, gf will be in mo basis)
            one of ao_orbs or mo_orbs must be None
            if both ao_orbs and mo_orbs are None, gf will be in mo basis
        gmres_tol : conjugate gradient (min res) conv tol (if too low will be extemely time-consuming)
        add_rem : '+' (addition) or '-' (removal) or '+-' (both)
        n_threads : number of threads (need parallel MKL library)
        memory : stack memory in bytes (default is 1 GB)
        verbose : 0 (quiet) 1 (per omega) 2 (per sweep) 3 (per orbital) 4 (per cg iteration)
        ignore_ecore : if True, set ecore to zero (should not affect GF)
        gs_bond_dims : np.ndarray of integers. Ground-State DMRG MPS bond dims for each sweep
        gs_noises : np.ndarray of float64. Ground-State DMRG noises for each sweep
        gs_tol : float64. Ground-State DMRG energy convergence.
        gs_n_steps : int. Ground-State DMRG max number of sweeps.
        cps_bond_dims : np.ndarray of integers. Compression MPS bond dims for each sweep
        cps_noises : np.ndarray of float64. Compression noises for each sweep
        cps_tol : float64. Compression energy convergence.
        cps_n_steps : int. Compression max number of sweeps.
        gf_bond_dims : np.ndarray of integers. Green's function MPS bond dims for each sweep
        gf_noises : np.ndarray of float64. Green's function noises for each sweep
        gf_tol : float64. Green's function Im GF (i, i) convergence.
        gf_n_steps : int. Green's function max number of sweeps.
        scratch : scratch folder for temporary files.
        lowdin : if True, will use lowdin orbitals instead of molecular orbitals
        diag_only : if True, only calculate diagonal GF elements.
        alpha : bool. alpha spin or beta spin (not used if SU2)
        occs : list(double) or None
            if occs = None, use FCI init MPS (default)
            if occs = list(double), use occ init MPS
            either
                (A) len(occs) == n_sites (RHF or UHF, 0 <= occ <= 2)
             or (B) len(occs) == n_sites * 2 (UHF, 0 <= occ <= 1)
                    order: 0a, 0b, 1a, 1b, ...
        bias : float64, effective when occs is not None
            bias = 0.0: HF occ (e.g. occs => 2 2 2 ... 0 0 0)
            bias = 1.0: no bias (e.g. occs => 1.7 1.8 ... 0.1 0.05)
            bias = +inf: fully random (e.g. occs => 1 1 1 ... 1 1 1)

            increase bias if you want an init MPS with larger bond dim

            0 <= occ <= 2
            biased occ = 1 + (occ - 1) ** bias    (if occ > 1)
                        = 1 - (1 - occ) ** bias    (if occ < 1)
        cutoff : float64, lowest kept density matrix eigen value (default 1E-14)
            use smaller cutoff (e.g. 1E-20) when you need super accurate ground state energy
            noise smaller than cutoff will have no effect.
        load_dir : if not None, skip ground-state calculation and load previous results
        save_dir : if not None, save results to a separate dir after ground-state calculation
            Note: one of load_dir and save_dir must be None
        mpi : if not None, MPI is used
        mo_basis: if False, will use Hamiltonian in AO basis
        pdm_return: if False, not return 1-pdm 
        reorder_method: None or 'fielder' or 'gaopt'
        n_off_diag_cg: limit number of cg for off-diagonal GF matrix elements
            if zero, no sweep is performed for off-diagonal GF matrix elements

    Returns:
        gfmat : np.ndarray of dims (len(mo_orbs), len(mo_orbs), len(freqs)) (complex)
            GF matrix in the MO basis (for selected mo_orbs).
    '''
    from pyscf import lib, lo, symm, ao2mo
    assert load_dir is None or save_dir is None
    assert ao_orbs is None or mo_orbs is None

    if mpi is not None:
        mpi = MPI

    mol = mf.mol

    pg = mol.symmetry.lower()
    if pg == 'd2h':
        fcidump_sym = ["Ag", "B3u", "B2u", "B1g", "B1u", "B2g", "B3g", "Au"]
    elif pg == 'c1':
        fcidump_sym = ["A"]
    else:
        raise GFDMRGError("Point group %d not supported yet!" % pg)

    if lowdin:
        mo_coeff = lo.orth.lowdin(mol.intor('cint1e_ovlp_sph'))
    else:
        mo_coeff = mf.mo_coeff

    is_uhf = isinstance(mo_coeff, tuple) or mo_coeff.ndim == 3

    if not is_uhf:

        if mo_basis:
            n_mo = mo_coeff.shape[1]

            # orb_sym_str = symm.label_orb_symm(
            #    mol, mol.irrep_name, mol.symm_orb, mo_coeff)
            #orb_sym = np.array([fcidump_sym.index(i) + 1 for i in orb_sym_str])
            orb_sym = [1] * n_mo

            h1e = mo_coeff.T @ mf.get_hcore() @ mo_coeff
            g2e = ao2mo.restore(8, ao2mo.kernel(mf._eri, mo_coeff), n_mo)
            ecore = mol.energy_nuc()
            if ignore_ecore:
                ecore = 0
            na = nb = mol.nelectron // 2
        else:
            n_mo = mo_coeff.shape[1]

            # orb_sym_str = symm.label_orb_symm(
            #    mol, mol.irrep_name, mol.symm_orb, np.eye(n_mo))
            #orb_sym = np.array([fcidump_sym.index(i) + 1 for i in orb_sym_str])
            orb_sym = [1] * n_mo

            h1e = mf.get_hcore()
            g2e = mf._eri
            ecore = mol.energy_nuc()
            if ignore_ecore:
                ecore = 0
            na = nb = mol.nelectron // 2

    else:

        if mo_basis:
            mo_coeff_a, mo_coeff_b = mo_coeff[0], mo_coeff[1]
            n_mo = mo_coeff_b.shape[1]

            # orb_sym_str_a = symm.label_orb_symm(
            #    mol, mol.irrep_name, mol.symm_orb, mo_coeff_a)
            #orb_sym_a = np.array([fcidump_sym.index(i) + 1 for i in orb_sym_str_a])

            #orb_sym = orb_sym_a
            orb_sym = [1] * n_mo

            h1ea = mo_coeff_a.T @ mf.get_hcore() @ mo_coeff_a
            h1eb = mo_coeff_b.T @ mf.get_hcore() @ mo_coeff_b
            g2eaa = ao2mo.restore(8, ao2mo.kernel(mf._eri, mo_coeff_a), n_mo)
            g2ebb = ao2mo.restore(8, ao2mo.kernel(mf._eri, mo_coeff_b), n_mo)
            g2eab = ao2mo.kernel(
                mf._eri, [mo_coeff_a, mo_coeff_a, mo_coeff_b, mo_coeff_b])
            h1e = (h1ea, h1eb)
            g2e = (g2eaa, g2ebb, g2eab)
            ecore = mol.energy_nuc()
            if ignore_ecore:
                ecore = 0
            na, nb = mol.nelectron

        else:
            mo_coeff_a, mo_coeff_b = mo_coeff[0], mo_coeff[1]
            n_mo = mo_coeff_b.shape[1]

            # orb_sym_str_a = symm.label_orb_symm(
            #    mol, mol.irrep_name, mol.symm_orb, mo_coeff_a)
            #orb_sym_a = np.array([fcidump_sym.index(i) + 1 for i in orb_sym_str_a])

            #orb_sym = orb_sym_a
            orb_sym = [1] * n_mo

            h1ea = mf.get_hcore()
            h1eb = h1ea.copy()
            g2eaa = mf._eri
            g2ebb = g2eaa.copy()
            g2eab = lib.unpack_tril(g2eaa)

            h1e = (h1ea, h1eb)
            g2e = (g2eaa, g2ebb, g2eab)
            ecore = mol.energy_nuc()
            if ignore_ecore:
                ecore = 0
            na, nb = mol.nelectron


    dmrg = GFDMRG(scratch=scratch, memory=memory,
                  verbose=verbose, omp_threads=n_threads, mpi=mpi)

    if load_dir is None:

        re_idx = orbital_reorder(h1e, g2e, method=reorder_method)
        if mpi is not None:
            from mpi4py import MPI as MPIPY
            comm = MPIPY.COMM_WORLD
            re_idx = comm.bcast(re_idx, root=0)

        save_fd = save_dir + "/GS_FCIDUMP" if save_dir is not None else None
        dmrg.init_hamiltonian(pg, n_sites=n_mo, n_elec=na + nb, twos=na - nb, isym=1,
                              orb_sym=orb_sym, e_core=ecore, h1e=h1e, g2e=g2e, idx=re_idx,
                              save_fcidump=save_fd)
        dmrg.dmrg(bond_dims=gs_bond_dims, noises=gs_noises,
                  n_steps=gs_n_steps, conv_tol=gs_tol, occs=occs, bias=bias, cutoff=cutoff)
        if save_dir is not None:
            _print('saving ground state ...')
            dmrg.save_gs_mps(save_dir)
            if mpi.rank == 0:
                np.save(save_dir + "/reorder.npy", re_idx)
    else:
        dmrg.init_hamiltonian_fcidump(pg, load_dir + "/GS_FCIDUMP")
        re_idx = np.load(load_dir + "/reorder.npy")
        dmrg.idx = re_idx
        dmrg.ridx = np.argsort(re_idx)
        _print('loading ground state ...')
        dmrg.load_gs_mps(load_dir)

    _print('reorder method = %r reorder = %r' % (reorder_method, re_idx))

    if mo_orbs is None and ao_orbs is None:
        mo_orbs = range(n_mo)
        mo_orbs = np.argsort(re_idx)[np.array(mo_orbs)]

    pdm = dmrg.get_one_pdm()
    idxs = mo_orbs if ao_orbs is None else ao_orbs
    if not is_uhf:
        gf = np.zeros((len(idxs), len(idxs), len(freqs)), dtype=complex)
    else:
        gf_beta = np.zeros((len(idxs), len(idxs), len(freqs)), dtype=complex)
    for iw in range(len(delta)):
        gf_tmp = 0
        for addit in [x == '+' for x in add_rem]:
            # only calculate alpha spin
            gf_tmp += dmrg.greens_function(gf_bond_dims, gf_noises, gmres_tol, gf_tol, gf_n_steps,
                                           cps_bond_dims, cps_noises, cps_tol, cps_n_steps,
                                           idxs=mo_orbs if ao_orbs is None else ao_orbs,
                                           mo_coeff=None if ao_orbs is None or not mo_basis else (mo_coeff[0] if is_uhf else mo_coeff),
                                           eta=delta[iw], freqs=np.array([freqs[iw]]), addition=addit, diag_only=diag_only,
                                           alpha=alpha, n_off_diag_cg=n_off_diag_cg)
        gf[:, :, iw] = gf_tmp[:, :, 0]

        if is_uhf:
            gf_beta_tmp = 0
            for addit in [x == '+' for x in add_rem]:
                # calculate beta spin
                gf_beta_tmp += dmrg.greens_function(gf_bond_dims, gf_noises, gmres_tol, gf_tol, gf_n_steps,
                                                    cps_bond_dims, cps_noises, cps_tol, cps_n_steps,
                                                    idxs=mo_orbs if ao_orbs is None else ao_orbs,
                                                    mo_coeff=None if ao_orbs is None or not mo_basis else (mo_coeff[1] if is_uhf else mo_coeff),
                                                    eta=delta[iw], freqs=np.array([freqs[iw]]), addition=addit, diag_only=diag_only,
                                                    alpha=False, n_off_diag_cg=n_off_diag_cg)
            gf_beta[:, :, iw] = gf_beta_tmp[:, :, 0]

    if is_uhf:
        gf = np.asarray((gf, gf_beta))

    del dmrg

    if pdm_return:
        return pdm, gf
    else:
        return gf


if __name__ == "__main__":

    # parameters
    n_threads = 14
    hf_type = "RHF"  # RHF or UHF
    mpg = 'c1'  # point group: d2h or c1
    scratch = './tmp'
    save_dir = './gs_mps'
    load_dir = None
    lowdin = False
    do_ccsd = True

    import os
    if MPI.rank == 0:
        if not os.path.isdir(scratch):
            os.mkdir(scratch)
        if save_dir is not None:
            if not os.path.isdir(save_dir):
                os.mkdir(save_dir)
    MPI.barrier()
    os.environ['TMPDIR'] = scratch

    from pyscf import gto, scf

    # H chain
    N = 10
    BOHR = 0.52917721092  # Angstroms
    R = 1.8 * BOHR
    mol = gto.M(atom=[['H', (0, 0, i * R)] for i in range(N)],
                basis='sto6g', verbose=0, symmetry=mpg)
    pg = mol.symmetry.lower()

    if hf_type == "RHF":
        mf = scf.RHF(mol)
        ener = mf.kernel()
    elif hf_type == "UHF":
        assert SpinLabel == SZ
        mf = scf.UHF(mol)
        ener = mf.kernel()

    _print("SCF Energy = %20.15f" % ener)
    _print(("NON-" if SpinLabel == SZ else "") + "SPIN-ADAPTED")

    if do_ccsd:
        from pyscf import cc
        mcc = cc.CCSD(mf)
        mcc.kernel()
        dmmo = mcc.make_rdm1()
        if hf_type == "RHF":
            # RHF: 0 <= occ <= 2, len = n_mo
            occs = np.diag(dmmo)
        else:
            # UHF: 0 <= occ <= 1, order: 0a, 0b, 1a, 1b, ..., len = n_mo * 2
            occs = np.array(
                [i for j in zip(np.diag(dmmo[0]), np.diag(dmmo[1])) for i in j])
        _print('OCCS = ', occs)
    else:
        occs = None

    if MPI is not None:
        from mpi4py import MPI as MPIPY
        comm = MPIPY.COMM_WORLD
        occs = comm.bcast(occs, root=0)
        ener = comm.bcast(ener, root=0)
        mf.mo_coeff = comm.bcast(mf.mo_coeff, root=0)

    if lowdin:

        eta = 0.005
        freqs = np.arange(-0.8, -0.2, 0.01)
        mo_orbs = [4]
        t = time.perf_counter()
        pdm, gfmat = dmrg_mo_gf(mf, freqs=freqs, delta=eta, mo_orbs=mo_orbs, scratch=scratch, add_rem='-',
                                gf_bond_dims=[150], gf_noises=[1E-3, 5E-4], gf_tol=1E-4,
                                gmres_tol=1E-8, lowdin=True, ignore_ecore=False, n_threads=n_threads, mpi=True)

        _print(gfmat)  # alpha only

        # alpha + beta
        pdos = (-2 / np.pi) * gfmat.imag.trace(axis1=0, axis2=1)
        _print("PDOS = ", pdos)
        _print("TIME = ", time.perf_counter() - t)

        import matplotlib.pyplot as plt

        plt.plot(freqs, pdos, 'o-', markersize=2)
        plt.xlabel('Frequency $\\omega$ (a.u.)')
        plt.ylabel('LDOS (a.u.)')
        plt.savefig('gf-figure.png', dpi=600)

    else:

        eta = [0.005]
        freqs = [-0.2]
        t = time.perf_counter()
        pdm, gfmat = dmrg_mo_gf(mf, freqs=freqs, delta=eta, mo_orbs=None, scratch=scratch, add_rem='+-',
                                gs_bond_dims=[500], gs_noises=[1E-7, 1E-8, 1E-10, 0], gs_tol=1E-14, gs_n_steps=30,
                                cps_bond_dims=[500], cps_noises=[0], cps_tol=1E-14, cps_n_steps=30,
                                gf_bond_dims=[500], gf_noises=[1E-7, 1E-8, 1E-10, 0], gf_tol=1E-8,
                                gmres_tol=1E-20, lowdin=False, ignore_ecore=False, alpha=False, verbose=3,
                                n_threads=n_threads, occs=None, bias=1.0, save_dir=save_dir, load_dir=load_dir,
                                n_off_diag_cg=-1,mpi=True)
        xgfmat = np.einsum('ip,pqr,jq->ijr', mf.mo_coeff, gfmat, mf.mo_coeff)
        _print("MO to AO method = ", xgfmat)

        pdm, gfmat = dmrg_mo_gf(mf, freqs=freqs, delta=eta, mo_orbs=None, scratch=scratch, add_rem='+-',
                                gs_bond_dims=[500], gs_noises=[1E-7, 1E-8, 1E-10, 0], gs_tol=1E-14, gs_n_steps=30,
                                cps_bond_dims=[500], cps_noises=[0], cps_tol=1E-14, cps_n_steps=30,
                                gf_bond_dims=[500], gf_noises=[1E-7, 1E-8, 1E-10, 0], gf_tol=1E-8,
                                gmres_tol=1E-20, lowdin=False, ignore_ecore=False, alpha=False, verbose=3,
                                ao_orbs=range(N),
                                n_threads=n_threads, occs=None, bias=1.0, save_dir=save_dir, load_dir=load_dir,mpi=True)
        _print("AO IN MO method = ", gfmat)

        _print('diff = ', gfmat - xgfmat)
        _print(np.linalg.norm(gfmat - xgfmat))




#  block2: Efficient MPO implementation of quantum chemistry DMRG
#  Copyright (C) 2020-2021 Huanchen Zhai <hczhai@caltech.edu>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#

"""
Low-Temperature (LT)-dmrg.
Multi-Target State-Averaged Excited-State Approach;
using pyscf and block2.

Author: Huanchen Zhai, Jun 21, 2020
"""

from block2 import init_memory, release_memory, Threading, ThreadingTypes, Global
from block2 import VectorUInt8, VectorUBond, VectorInt, VectorDouble, PointGroup
from block2 import Random, FCIDUMP, QCTypes, SeqTypes
from block2 import SU2, SZ, get_partition_weights, OpNamesSet, OpNames
import numpy as np
import time

# Set spin-adapted or non-spin-adapted here
# SpinLabel = SU2
SpinLabel = SZ

if SpinLabel == SU2:
    from block2 import VectorSU2 as VectorSL
    from block2.su2 import HamiltonianQC, MultiMPSInfo, MultiMPS
    from block2.su2 import PDM1MPOQC, NPC1MPOQC, SimplifiedMPO, Rule, RuleQC, MPOQC
    from block2.su2 import MovingEnvironment, Expect, DMRG
else:
    from block2 import VectorSZ as VectorSL
    from block2.sz import HamiltonianQC, MultiMPSInfo, MultiMPS
    from block2.sz import PDM1MPOQC, PDM2MPOQC, NPC1MPOQC, SimplifiedMPO, Rule, RuleQC, MPOQC
    from block2.sz import MovingEnvironment, Expect, DMRG


class LTDMRGError(Exception):
    pass


class LTDMRG:
    """
    Low-temperature DMRG for molecules.
    """

    def __init__(self, scratch='./nodex', memory=1 * 1E9, omp_threads=2, verbose=2):

        Random.rand_seed(0)
        init_memory(isize=int(memory * 0.1),
                    dsize=int(memory * 0.9), save_dir=scratch)
        Global.threading = Threading(
            ThreadingTypes.OperatorBatchedGEMM | ThreadingTypes.Global, omp_threads, omp_threads, 1)
        Global.threading.seq_type = SeqTypes.Tasked
        Global.frame.use_main_stack = False
        self.fcidump = None
        self.hamil = None
        self.verbose = verbose
        self.nroots = 10

    def init_hamiltonian_fcidump(self, pg, filename):
        assert self.fcidump is None
        self.fcidump = FCIDUMP()
        self.fcidump.read(filename)
        self.orb_sym = VectorUInt8(
            map(PointGroup.swap_d2h, self.fcidump.orb_sym))

        vacuum = SpinLabel(0)
        self.targets = VectorSL([SpinLabel(self.fcidump.n_elec, self.fcidump.twos,
                                           PointGroup.swap_d2h(self.fcidump.isym))])
        self.n_sites = self.fcidump.n_sites

        self.hamil = HamiltonianQC(
            vacuum, self.n_sites, self.orb_sym, self.fcidump)
        assert pg in ["d2h", "c1"]

    def init_hamiltonian(self, pg, n_sites, n_elec, twos, isym, orb_sym, e_core,
                         h1e, g2e, tol=1E-13, save_fcidump=None):
        assert self.fcidump is None
        self.fcidump = FCIDUMP()
        if not isinstance(h1e, tuple):
            mh1e = np.zeros((n_sites * (n_sites + 1) // 2))
            k = 0
            for i in range(0, n_sites):
                for j in range(0, i + 1):
                    assert abs(h1e[i, j] - h1e[j, i]) < tol
                    mh1e[k] = h1e[i, j]
                    k += 1
            mg2e = g2e.flatten()
            mh1e[np.abs(mh1e) < tol] = 0.0
            mg2e[np.abs(mg2e) < tol] = 0.0
            self.fcidump.initialize_su2(
                n_sites, n_elec, twos, isym, e_core, mh1e, mg2e)
        else:
            assert SpinLabel == SZ
            assert isinstance(h1e, tuple) and len(h1e) == 2
            assert isinstance(g2e, tuple) and len(g2e) == 3
            mh1e_a = np.zeros((n_sites * (n_sites + 1) // 2))
            mh1e_b = np.zeros((n_sites * (n_sites + 1) // 2))
            mh1e = (mh1e_a, mh1e_b)
            for xmh1e, xh1e in zip(mh1e, h1e):
                k = 0
                for i in range(0, n_sites):
                    for j in range(0, i + 1):
                        assert abs(xh1e[i, j] - xh1e[j, i]) < tol
                        xmh1e[k] = xh1e[i, j]
                        k += 1
                xmh1e[np.abs(xmh1e) < tol] = 0.0
            mg2e = tuple(xg2e.flatten() for xg2e in g2e)
            for xmg2e in mg2e:
                xmg2e[np.abs(xmg2e) < tol] = 0.0
            self.fcidump.initialize_sz(
                n_sites, n_elec, twos, isym, e_core, mh1e, mg2e)
        self.orb_sym = VectorUInt8(map(PointGroup.swap_d2h, orb_sym))

        vacuum = SpinLabel(0)
        self.targets = VectorSL(
            [SpinLabel(n_elec, twos, PointGroup.swap_d2h(isym))])
        self.n_sites = n_sites

        self.hamil = HamiltonianQC(
            vacuum, self.n_sites, self.orb_sym, self.fcidump)

        if save_fcidump is not None:
            self.fcidump.orb_sym = VectorUInt8(orb_sym)
            self.fcidump.write(save_fcidump)
        assert pg in ["d2h", "c1"]

    # State-averaged DMRG for multipile targets and multiple roots
    def dmrg(self, mu, bond_dims, noises, n_steps=30, conv_tol=1E-7):

        if self.verbose >= 2:
            print('>>> START DMRG <<<')
        t = time.perf_counter()

        self.hamil.mu = mu

        # MultiMPSInfo
        mps_info = MultiMPSInfo(self.n_sites, self.hamil.vacuum,
                                self.targets, self.hamil.basis)
        mps_info.tag = "FINAL"
        mps_info.set_bond_dimension(bond_dims[0])

        # MultiMPS
        mps = MultiMPS(self.n_sites, self.n_sites - 2, 2, self.nroots)
        mps.initialize(mps_info)
        mps.random_canonicalize()

        mps.save_mutable()
        mps.deallocate()
        mps_info.save_mutable()
        mps_info.deallocate_mutable()

        # MPO
        tx = time.perf_counter()
        mpo = MPOQC(self.hamil, QCTypes.Conventional)
        mpo = SimplifiedMPO(mpo, RuleQC(), True, True,
                            OpNamesSet((OpNames.R, OpNames.RD)))
        if self.verbose >= 2:
            print('MPO time = ', time.perf_counter() - tx)

        # DMRG
        me = MovingEnvironment(mpo, mps, mps, "DMRG")
        me.delayed_contraction = OpNamesSet.normal_ops()
        me.cached_contraction = True
        tx = time.perf_counter()
        me.init_environments(self.verbose >= 3)
        if self.verbose >= 2:
            print('DMRG INIT time = ', time.perf_counter() - tx)
        dmrg = DMRG(me, VectorUBond(bond_dims), VectorDouble(noises))
        dmrg.davidson_max_iter = 1000 * self.nroots
        dmrg.iprint = self.verbose
        dmrg.solve(n_steps, mps.center == 0)

        self.energies = dmrg.energies[-1]
        self.quanta = dmrg.mps_quanta[-1]
        self.multiplicities = VectorInt(
            [qs[0][0].multiplicity for qs in self.quanta])

        self.bond_dim = bond_dims[-1]
        mps.save_data()
        mpo.deallocate()
        mps_info.deallocate()

        if self.verbose >= 2:
            print('>>> COMPLETE DMRG | Time = %.2f <<<' %
                  (time.perf_counter() - t))

    # particle number correlation
    # return value (SU2):
    #     npc[0, :, :] -> <(N_{i,alpha}+N_{i,beta}) (N_{j,alpha}+N_{j,beta})>
    # return value (SZ):
    #     npc[0, :, :] -> <N_{i,alpha} N_{j,alpha}>
    #     npc[1, :, :] -> <N_{i,alpha}  N_{j,beta}>
    #     npc[2, :, :] -> < N_{i,beta} N_{j,alpha}>
    #     npc[3, :, :] -> < N_{i,beta}  N_{j,beta}>
    def get_one_npc(self, beta, ridx=None):
        if self.verbose >= 2:
            print('>>> START one-npc <<<')
        t = time.perf_counter()

        self.hamil.mu = 0.0
        assert isinstance(beta, float)

        # MultiMPSInfo
        mps_info = MultiMPSInfo(self.n_sites, self.hamil.vacuum,
                                self.targets, self.hamil.basis)
        mps_info.tag = "FINAL"
        mps_info.load_mutable()

        # MultiMPS (final)
        mps = MultiMPS(mps_info)
        mps.load_data()

        # 1NPC MPO
        pmpo = NPC1MPOQC(self.hamil)
        pmpo = SimplifiedMPO(pmpo, Rule())

        dms = []

        for iroot in range(mps.nroots):
            if self.verbose >= 2:
                print('>>> root = %3d / %3d <<<' % (iroot, mps.nroots))

            smps_info = MultiMPSInfo(self.n_sites, self.hamil.vacuum,
                                     self.targets, self.hamil.basis)
            smps_info.tag = "1NPC"
            for i in range(0, smps_info.n_sites + 1):
                smps_info.left_dims[i] = mps_info.left_dims[i]
                smps_info.right_dims[i] = mps_info.right_dims[i]
            smps_info.save_mutable()

            mps.load_mutable()

            smps = MultiMPS(smps_info)
            smps.n_sites = mps.n_sites
            smps.center = mps.center
            smps.dot = mps.dot
            smps.canonical_form = '' + mps.canonical_form
            smps.tensors = mps.tensors[:]
            smps.wfns = mps.wfns[iroot:iroot+1]
            smps.weights = mps.weights[iroot:iroot+1]
            smps.nroots = 1
            smps.save_mutable()

            mps.deallocate()

            # 1NPC
            pme = MovingEnvironment(pmpo, smps, smps, "1NPC")
            pme.init_environments(self.verbose >= 3)
            expect = Expect(pme, self.bond_dim, self.bond_dim)
            expect.iprint = self.verbose
            expect.solve(True, smps.center == 0)

            if SpinLabel == SU2:
                dmr = expect.get_1npc_spatial(0, self.n_sites)
                dm = np.array(dmr).copy()
            else:
                dmr = expect.get_1npc(0, self.n_sites)
                dm = np.array(dmr).copy()
                dm = dm.reshape((self.n_sites, 2, self.n_sites, 2))
                dm = np.transpose(dm, (0, 2, 1, 3))

            dmr.deallocate()
            smps_info.deallocate()

            if ridx is not None:
                dm[:, :] = dm[ridx, :][:, ridx]

            if SpinLabel == SU2:
                dm = dm[None, :, :]
            else:
                dm = np.concatenate([dm[None, :, :, 0, 0], dm[None, :, :, 0, 1],
                                     dm[None, :, :, 1, 0], dm[None, :, :, 1, 1]], axis=0)

            dms.append(dm)

        pmpo.deallocate()
        mps_info.deallocate_mutable()
        mps_info.deallocate()

        pws = get_partition_weights(beta, self.energies, self.multiplicities)
        dm = np.add.reduce([w * dm for w, dm in zip(pws, dms)])

        if self.verbose >= 2:
            print('>>> COMPLETE one-npc | Time = %.2f <<<' %
                  (time.perf_counter() - t))

        return dm

    # one-particle density matrix
    # return value:
    #     pdm[0, :, :] -> <AD_{i,alpha} A_{j,alpha}>
    #     pdm[1, :, :] -> < AD_{i,beta}  A_{j,beta}>
    def get_one_pdm(self, beta, ridx=None):
        if self.verbose >= 2:
            print('>>> START one-pdm <<<')
        t = time.perf_counter()

        self.hamil.mu = 0.0
        assert isinstance(beta, float)

        # MultiMPSInfo
        mps_info = MultiMPSInfo(self.n_sites, self.hamil.vacuum,
                                self.targets, self.hamil.basis)
        mps_info.tag = "FINAL"
        mps_info.load_mutable()

        # MultiMPS (final)
        mps = MultiMPS(mps_info)
        mps.load_data()

        # 1PDM MPO
        pmpo = PDM1MPOQC(self.hamil)
        pmpo = SimplifiedMPO(pmpo, RuleQC())

        dms = []

        for iroot in range(mps.nroots):
            if self.verbose >= 2:
                print('>>> root = %3d / %3d <<<' % (iroot, mps.nroots))

            smps_info = MultiMPSInfo(self.n_sites, self.hamil.vacuum,
                                     self.targets, self.hamil.basis)
            smps_info.tag = "1PDM"
            for i in range(0, smps_info.n_sites + 1):
                smps_info.left_dims[i] = mps_info.left_dims[i]
                smps_info.right_dims[i] = mps_info.right_dims[i]
            smps_info.save_mutable()

            mps.load_mutable()

            smps = MultiMPS(smps_info)
            smps.n_sites = mps.n_sites
            smps.center = mps.center
            smps.dot = mps.dot
            smps.canonical_form = '' + mps.canonical_form
            smps.tensors = mps.tensors[:]
            smps.wfns = mps.wfns[iroot:iroot+1]
            smps.weights = mps.weights[iroot:iroot+1]
            smps.nroots = 1
            smps.save_mutable()

            mps.deallocate()

            # 1PDM
            pme = MovingEnvironment(pmpo, smps, smps, "1PDM")
            pme.init_environments(self.verbose >= 3)
            expect = Expect(pme, self.bond_dim, self.bond_dim)
            expect.iprint = self.verbose
            expect.solve(True, smps.center == 0)

            if SpinLabel == SU2:
                dmr = expect.get_1pdm_spatial(self.n_sites)
                dm = np.array(dmr).copy()
            else:
                dmr = expect.get_1pdm(self.n_sites)
                dm = np.array(dmr).copy()
                dm = dm.reshape((self.n_sites, 2, self.n_sites, 2))
                dm = np.transpose(dm, (0, 2, 1, 3))

            dmr.deallocate()
            smps_info.deallocate()

            if ridx is not None:
                dm[:, :] = dm[ridx, :][:, ridx]

            if SpinLabel == SU2:
                dm = np.concatenate(
                    [dm[None, :, :], dm[None, :, :]], axis=0) / 2
            else:
                dm = np.concatenate(
                    [dm[None, :, :, 0, 0], dm[None, :, :, 1, 1]], axis=0)

            dms.append(dm)

        pmpo.deallocate()
        mps_info.deallocate_mutable()
        mps_info.deallocate()

        pws = get_partition_weights(beta, self.energies, self.multiplicities)
        dm = np.add.reduce([w * dm for w, dm in zip(pws, dms)])

        if self.verbose >= 2:
            print('>>> COMPLETE one-pdm | Time = %.2f <<<' %
                  (time.perf_counter() - t))

        return dm

    # two-particle density matrix (SZ only)
    # return value:
    #     pdm[0, i, j, k, l] -> <AD_{i,alpha} AD_{j,alpha} A_{k,alpha} A_{l,alpha}>
    #     pdm[1, i, j, k, l] -> <AD_{i,alpha} AD_{j, beta} A_{k, beta} A_{l,alpha}>
    #     pdm[2, i, j, k, l] -> <AD_{i, beta} AD_{j, beta} A_{k, beta} A_{l, beta}>
    def get_two_pdm(self, beta, ridx=None):
        if self.verbose >= 2:
            print('>>> START two-pdm <<<')
        t = time.perf_counter()

        self.hamil.mu = 0.0

        # only support SZ
        assert SpinLabel == SZ

        assert isinstance(beta, float)

        # MultiMPSInfo
        mps_info = MultiMPSInfo(self.n_sites, self.hamil.vacuum,
                                self.targets, self.hamil.basis)
        mps_info.tag = "FINAL"
        mps_info.load_mutable()

        # MultiMPS (final)
        mps = MultiMPS(mps_info)
        mps.load_data()

        # 2PDM MPO
        pmpo = PDM2MPOQC(self.hamil, mask=PDM2MPOQC.s_minimal)
        pmpo = SimplifiedMPO(pmpo, RuleQC())

        dms = []

        for iroot in range(mps.nroots):
            if self.verbose >= 2:
                print('>>> root = %3d / %3d <<<' % (iroot, mps.nroots))

            smps_info = MultiMPSInfo(self.n_sites, self.hamil.vacuum,
                                     self.targets, self.hamil.basis)
            smps_info.tag = "2PDM"
            for i in range(0, smps_info.n_sites + 1):
                smps_info.left_dims[i] = mps_info.left_dims[i]
                smps_info.right_dims[i] = mps_info.right_dims[i]
            smps_info.save_mutable()

            mps.load_mutable()

            smps = MultiMPS(smps_info)
            smps.n_sites = mps.n_sites
            smps.center = mps.center
            smps.dot = mps.dot
            smps.canonical_form = '' + mps.canonical_form
            smps.tensors = mps.tensors[:]
            smps.wfns = mps.wfns[iroot:iroot+1]
            smps.weights = mps.weights[iroot:iroot+1]
            smps.nroots = 1
            smps.save_mutable()

            mps.deallocate()

            # 2PDM
            pme = MovingEnvironment(pmpo, smps, smps, "2PDM")
            pme.init_environments(self.verbose >= 3)
            expect = Expect(pme, self.bond_dim, self.bond_dim)
            expect.iprint = self.verbose
            expect.solve(True, smps.center == 0)

            dmr = expect.get_2pdm(self.n_sites)
            dm = np.array(dmr, copy=True)
            dm = dm.reshape((self.n_sites, 2, self.n_sites, 2,
                             self.n_sites, 2, self.n_sites, 2))
            dm = np.transpose(dm, (0, 2, 4, 6, 1, 3, 5, 7))

            smps_info.deallocate()

            if ridx is not None:
                dm[:, :, :, :] = dm[ridx, :, :, :] \
                    [:, ridx, :, :][:, :, ridx, :][:, :, :, ridx]

            dm = np.concatenate([dm[None, :, :, :, :, 0, 0, 0, 0], dm[None, :, :, :, :, 0, 1, 1, 0],
                                 dm[None, :, :, :, :, 1, 1, 1, 1]], axis=0)

            dms.append(dm)

        pmpo.deallocate()
        mps_info.deallocate_mutable()
        mps_info.deallocate()

        pws = get_partition_weights(beta, self.energies, self.multiplicities)
        dm = np.add.reduce([w * dm for w, dm in zip(pws, dms)])

        if self.verbose >= 2:
            print('>>> COMPLETE two-pdm | Time = %.2f <<<' %
                  (time.perf_counter() - t))

        return dm

    def __del__(self):
        if self.hamil is not None:
            self.hamil.deallocate()
        if self.fcidump is not None:
            self.fcidump.deallocate()
        release_memory()


if __name__ == "__main__":

    # parameters
    bond_dims = [250, 250, 500, 500, 500, 500, 750]
    noises = [1E-7, 1E-7, 1E-8, 1E-8, 1E-8, 1E-9, 1E-9, 1E-9, 0]
    sweep_tol = 1E-8
    max_dmrg_steps = 30

    nroots = 30
    beta = 20.0
    mu = -1.0
    n_threads = 8
    hf_type = "RHF"
    mpg = 'd2h'
    pg_reorder = True
    scratch = './nodex'
    scratch = '/scratch/local/hczhai/lt-hchain'

    import os
    if not os.path.isdir(scratch):
        os.mkdir(scratch)
    os.environ['TMPDIR'] = scratch

    from pyscf import gto, scf, symm, ao2mo

    # H chain
    N = 8
    BOHR = 0.52917721092  # Angstroms
    R = 1.8 * BOHR
    mol = gto.M(atom=[['H', (i * R, 0, 0)] for i in range(N)],
                basis='sto6g', verbose=0, symmetry=mpg)
    pg = mol.symmetry.lower()
    nelec = mol.nelectron

    # Reorder
    if pg == 'd2h':
        fcidump_sym = ["Ag", "B3u", "B2u", "B1g", "B1u", "B2g", "B3g", "Au"]
        optimal_reorder = ["Ag", "B1u", "B3u",
                           "B2g", "B2u", "B3g", "B1g", "Au"]
    elif pg == 'c1':
        fcidump_sym = ["A"]
        optimal_reorder = ["A"]
    else:
        raise LTDMRGError("Point group %d not supported yet!" % pg)

    if hf_type == "RHF":
        # SCF
        m = scf.RHF(mol)
        m.kernel()
        mo_coeff = m.mo_coeff
        n_ao = mo_coeff.shape[0]
        n_mo = mo_coeff.shape[1]

        orb_sym_str = symm.label_orb_symm(
            mol, mol.irrep_name, mol.symm_orb, mo_coeff)
        orb_sym = np.array([fcidump_sym.index(i) + 1 for i in orb_sym_str])

        # Sort the orbitals by symmetry for more efficient DMRG
        if pg_reorder:
            idx = np.argsort([optimal_reorder.index(i) for i in orb_sym_str])
            orb_sym = orb_sym[idx]
            mo_coeff = mo_coeff[:, idx]
            ridx = np.argsort(idx)
        else:
            ridx = np.array(list(range(n_mo)), dtype=int)

        h1e = mo_coeff.T @ m.get_hcore() @ mo_coeff
        g2e = ao2mo.restore(8, ao2mo.kernel(mol, mo_coeff), n_mo)
        ecore = mol.energy_nuc()
        ecore = 0.0

    elif hf_type == "UHF":
        assert SpinLabel == SZ
        # SCF
        m = scf.UHF(mol)
        m.kernel()
        mo_coeff_a, mo_coeff_b = m.mo_coeff[0], m.mo_coeff[1]
        n_ao = mo_coeff_a.shape[0]
        n_mo = mo_coeff_b.shape[1]

        orb_sym_str_a = symm.label_orb_symm(
            mol, mol.irrep_name, mol.symm_orb, mo_coeff_a)
        orb_sym_str_b = symm.label_orb_symm(
            mol, mol.irrep_name, mol.symm_orb, mo_coeff_b)
        orb_sym_a = np.array([fcidump_sym.index(i) + 1 for i in orb_sym_str_a])
        orb_sym_b = np.array([fcidump_sym.index(i) + 1 for i in orb_sym_str_b])

        # Sort the orbitals by symmetry for more efficient DMRG
        if pg_reorder:
            idx_a = np.argsort([optimal_reorder.index(i)
                                for i in orb_sym_str_a])
            orb_sym_a = orb_sym_a[idx_a]
            mo_coeff_a = mo_coeff_a[:, idx_a]
            idx_b = np.argsort([optimal_reorder.index(i)
                                for i in orb_sym_str_b])
            orb_sym_b = orb_sym_b[idx_b]
            mo_coeff_b = mo_coeff_b[:, idx_b]
            assert np.allclose(idx_a, idx_b)
            assert np.allclose(orb_sym_a, orb_sym_b)
            orb_sym = orb_sym_a
            ridx = np.argsort(idx_a)
        else:
            orb_sym = orb_sym_a
            ridx = np.array(list(range(n_mo)), dtype=int)

        h1ea = mo_coeff_a.T @ m.get_hcore() @ mo_coeff_a
        h1eb = mo_coeff_b.T @ m.get_hcore() @ mo_coeff_b
        g2eaa = ao2mo.restore(8, ao2mo.kernel(mol, mo_coeff_a), n_mo)
        g2ebb = ao2mo.restore(8, ao2mo.kernel(mol, mo_coeff_b), n_mo)
        g2eab = ao2mo.kernel(
            mol, [mo_coeff_a, mo_coeff_a, mo_coeff_b, mo_coeff_b])
        h1e = (h1ea, h1eb)
        g2e = (g2eaa, g2ebb, g2eab)
        ecore = mol.energy_nuc()
        ecore = 0.0

    lt = LTDMRG(scratch=scratch, memory=2E9, verbose=2, omp_threads=n_threads)
    lt.init_hamiltonian(pg, n_sites=n_mo, n_elec=nelec, twos=0, isym=1,
                        orb_sym=orb_sym, e_core=ecore, h1e=h1e, g2e=g2e)

    # Set multiple targets
    lt.targets = VectorSL([])
    pgs = list(set(list(lt.orb_sym)))
    for xpg in pgs:
        for na in range(nelec // 2 - 3, nelec // 2 + 4):
            for nb in range(nelec // 2 - 3, nelec // 2 + 4):
                xn = na + nb
                xs2 = na - nb
                if SpinLabel == SU2 and xs2 < 0:
                    continue
                lt.targets.append(SpinLabel(xn, xs2, xpg))
    # Numebr of states
    lt.nroots = nroots
    print("Targets : LEN = ", len(lt.targets), lt.targets)
    print("Nroots = ", lt.nroots)

    lt.dmrg(mu, bond_dims, noises, n_steps=max_dmrg_steps, conv_tol=sweep_tol)

    partition_weights = get_partition_weights(
        beta, lt.energies, lt.multiplicities)

    print("Partition Function Weights (beta = %9.5f, mu = %9.5f):" % (beta, mu))
    for ii, (e, qs, pw) in enumerate(zip(lt.energies, lt.quanta, partition_weights)):
        print("   [State %3d] = %20.15f %20r %10.5g" % (ii, e, qs[0][0], pw))

    print("Multiplicities = ", lt.multiplicities)
    print("Average Energy = %20.15f" % sum([(e + mu * qs[0][0].n) * pw
                                            for e, qs, pw in zip(lt.energies, lt.quanta, partition_weights)]))
    print("Average Particle Number = %20.15f" % sum([qs[0][0].n * pw
                                                     for qs, pw in zip(lt.quanta, partition_weights)]))

    pdm1 = lt.get_one_pdm(beta, ridx)
    npc1 = lt.get_one_npc(beta, ridx)
    pdm2 = lt.get_two_pdm(beta, ridx)

    import pickle
    pickle.dump(pdm1, open('ltpdm1', "wb"))
    pickle.dump(pdm2, open('ltpdm2', "wb"))

    del lt




#  block2: Efficient MPO implementation of quantum chemistry DMRG
#  Copyright (C) 2020-2021 Huanchen Zhai <hczhai@caltech.edu>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#

"""
DMRG with state interaction spin-orbit (SISO) coupling
using pyscf and block2.

Author: Huanchen Zhai, Mar 22, 2021
"""

from block2 import SU2, SZ, Global, OpNamesSet, Threading, ThreadingTypes
from block2 import init_memory, release_memory, set_mkl_num_threads, SiteIndex
from block2 import VectorUInt8, PointGroup, FCIDUMP, QCTypes, SeqTypes, OpNames, Random
from block2 import VectorUBond, VectorDouble, NoiseTypes, DecompositionTypes, EquationTypes
import time
import numpy as np
import copy

# Set spin-adapted or non-spin-adapted here
SpinLabel = SU2
# SpinLabel = SZ

if SpinLabel == SU2:
    from block2 import VectorSU2 as VectorSL
    from block2.su2 import HamiltonianQC, SimplifiedMPO, Rule, RuleQC, MPOQC, CG
    from block2.su2 import MPSInfo, MPS, MovingEnvironment, DMRG, Linear, IdentityMPO
    from block2.su2 import OpElement, SiteMPO, NoTransposeRule, PDM1MPOQC, Expect
    from block2.su2 import MultiMPSInfo, MultiMPS, ParallelMPO, ParallelRuleQC, ParallelRuleNPDMQC
else:
    from block2 import VectorSZ as VectorSL
    from block2.sz import HamiltonianQC, SimplifiedMPO, Rule, RuleQC, MPOQC, CG
    from block2.sz import MPSInfo, MPS, MovingEnvironment, DMRG, Linear, IdentityMPO
    from block2.sz import OpElement, SiteMPO, NoTransposeRule, PDM1MPOQC, Expect
    from block2.sz import MultiMPSInfo, MultiMPS, ParallelMPO, ParallelRuleQC, ParallelRuleNPDMQC

try:
    if SpinLabel == SU2:
        from block2.su2 import MPICommunicator
    else:
        from block2.sz import MPICommunicator
    MPI = MPICommunicator()
    from mpi4py import MPI as PYMPI
    comm = PYMPI.COMM_WORLD

    def _print(*args, **kwargs):
        if MPI.rank == 0:
            print(*args, **kwargs)
except ImportError:
    MPI = None
    _print = print


class SIDMRGError(Exception):
    pass


class SIDMRG:
    """
    Multi-State DMRG for molecules.
    """

    def __init__(self, scratch='./nodex', memory=1 * 1E9, omp_threads=8, verbose=2,
                 print_statistics=False, mpi=None, dctr=True):
        """
        Memory is in bytes.
        verbose = 0 (quiet), 2 (per sweep), 3 (per iteration)
        """

        Random.rand_seed(0)
        init_memory(isize=int(memory * 0.1),
                    dsize=int(memory * 0.9), save_dir=scratch)
        Global.threading = Threading(
            ThreadingTypes.OperatorBatchedGEMM | ThreadingTypes.Global, omp_threads, omp_threads, 1)
        Global.threading.seq_type = SeqTypes.Simple
        Global.frame.load_buffering = False
        Global.frame.save_buffering = False
        Global.frame.use_main_stack = False
        self.fcidump = None
        self.hamil = None
        self.verbose = verbose
        self.scratch = scratch
        self.mpo_orig = None
        self.print_statistics = print_statistics
        self.mpi = mpi
        self.dctr = dctr

        if self.verbose >= 2:
            print(Global.frame)
            print(Global.threading)

        if mpi is not None:
            self.prule = ParallelRuleQC(mpi)
            self.pdmrule = ParallelRuleNPDMQC(mpi)
        else:
            self.prule = None
            self.pdmrule = None

    def init_hamiltonian_fcidump(self, pg, filename):
        """Read integrals from FCIDUMP file.
        pg : point group, pg = 'c1', 'c2v' or 'd2h'
        filename : FCIDUMP filename
        """
        if self.hamil is not None:
            self.hamil.deallocate()
        if self.fcidump is not None:
            self.fcidump.deallocate()
        if self.mpo_orig is not None:
            self.mpo_orig.deallocate()
        self.fcidump = FCIDUMP()
        self.fcidump.read(filename)
        swap_pg = getattr(PointGroup, "swap_" + pg)
        self.orb_sym = VectorUInt8(map(swap_pg, self.fcidump.orb_sym))

        vacuum = SpinLabel(0)
        self.target = SpinLabel(self.fcidump.n_elec, self.fcidump.twos,
                                swap_pg(self.fcidump.isym))
        self.n_sites = self.fcidump.n_sites

        self.hamil = HamiltonianQC(
            vacuum, self.n_sites, self.orb_sym, self.fcidump)
        assert pg in ["d2h", "c2v", "c1"]

    def init_hamiltonian(self, pg, n_sites, n_elec, twos, isym, orb_sym,
                         e_core, h1e, g2e, tol=1E-13, save_fcidump=None):
        """Initialize integrals using h1e, g2e, etc."""
        if self.hamil is not None:
            self.hamil.deallocate()
        if self.fcidump is not None:
            self.fcidump.deallocate()
        if self.mpo_orig is not None:
            self.mpo_orig.deallocate()
        self.fcidump = FCIDUMP()
        if not isinstance(h1e, tuple):
            mh1e = np.zeros((n_sites * (n_sites + 1) // 2))
            k = 0
            for i in range(0, n_sites):
                for j in range(0, i + 1):
                    assert abs(h1e[i, j] - h1e[j, i]) < tol
                    mh1e[k] = h1e[i, j]
                    k += 1
            mg2e = g2e.flatten()
            mh1e[np.abs(mh1e) < tol] = 0.0
            mg2e[np.abs(mg2e) < tol] = 0.0
            self.fcidump.initialize_su2(
                n_sites, n_elec, twos, isym, e_core, mh1e, mg2e)
        else:
            assert SpinLabel == SZ
            assert isinstance(h1e, tuple) and len(h1e) == 2
            assert isinstance(g2e, tuple) and len(g2e) == 3
            mh1e_a = np.zeros((n_sites * (n_sites + 1) // 2))
            mh1e_b = np.zeros((n_sites * (n_sites + 1) // 2))
            mh1e = (mh1e_a, mh1e_b)
            for xmh1e, xh1e in zip(mh1e, h1e):
                k = 0
                for i in range(0, n_sites):
                    for j in range(0, i + 1):
                        assert abs(xh1e[i, j] - xh1e[j, i]) < tol
                        xmh1e[k] = xh1e[i, j]
                        k += 1
                xmh1e[np.abs(xmh1e) < tol] = 0.0
            mg2e = tuple(xg2e.flatten() for xg2e in g2e)
            for xmg2e in mg2e:
                xmg2e[np.abs(xmg2e) < tol] = 0.0
            self.fcidump.initialize_sz(
                n_sites, n_elec, twos, isym, e_core, mh1e, mg2e)
        swap_pg = getattr(PointGroup, "swap_" + pg)
        self.orb_sym = VectorUInt8(map(swap_pg, orb_sym))

        vacuum = SpinLabel(0)
        self.target = SpinLabel(n_elec, twos, swap_pg(isym))
        self.n_sites = n_sites

        self.hamil = HamiltonianQC(
            vacuum, self.n_sites, self.orb_sym, self.fcidump)

        if save_fcidump is not None:
            self.fcidump.orb_sym = VectorUInt8(orb_sym)
            if self.mpi is None or self.mpi.rank == 0:
                self.fcidump.write(save_fcidump)
        assert pg in ["d2h", "c2v", "c1"]

    @staticmethod
    def fmt_size(i, suffix='B'):
        if i < 1000:
            return "%d %s" % (i, suffix)
        else:
            a = 1024
            for pf in "KMGTPEZY":
                p = 2
                for k in [10, 100, 1000]:
                    if i < k * a:
                        return "%%.%df %%s%%s" % p % (i / a, pf, suffix)
                    p -= 1
                a *= 1024
        return "??? " + suffix

    def dmrg(self, target, nroots, bond_dims, noises, dav_thrds, weights=None,
             tag='KET', occs=None, bias=1.0, n_steps=30, conv_tol=1E-7, cutoff=1E-14):
        """State-averaged multi-state DMRG.

        Args:
            target : quantum number of wavefunction
            nroots : number of states to solve
            weights : list(float) weight of each root
            tag : str; MPS tag
            bond_dims : list(int), MPS bond dims for each sweep
            noises : list(double), noise for each sweep
            dav_thrds : list(double), Davidson convergence for each sweep
            occs : list(double) or None
                if occs = None, use FCI init MPS
                if occs = list(double), use occ init MPS
            bias : double, effective when occs is not None
                bias = 0.0: HF occ (e.g. occs => 2 2 2 ... 0 0 0)
                bias = 1.0: no bias (e.g. occs => 1.7 1.8 ... 0.1 0.05)
                bias = +inf: fully random (e.g. occs => 1 1 1 ... 1 1 1)

                increase bias if you want an init MPS with larger bond dim

                0 <= occ <= 2
                biased occ = 1 + (occ - 1) ** bias    (if occ > 1)
                           = 1 - (1 - occ) ** bias    (if occ < 1)
            n_steps : int, maximal number of sweeps
            conv_tol : double, energy convergence
            cutoff : cutoff of density matrix eigenvalues (default is the same as StackBlock)
        """

        if self.verbose >= 2:
            print('>>> START State-Averaged DMRG <<<')
        t = time.perf_counter()

        if self.mpi is not None:
            self.mpi.barrier()

        # MPSInfo
        targets = VectorSL([target])
        _print('TARGETS = ', list(targets), flush=True)
        mps_info = MultiMPSInfo(
            self.n_sites, self.hamil.vacuum, targets, self.hamil.basis)
        mps_info.tag = tag
        if occs is None:
            if self.verbose >= 2:
                _print("Using FCI INIT MPS")
            mps_info.set_bond_dimension(bond_dims[0])
        else:
            if self.verbose >= 2:
                _print("Using occupation number INIT MPS")
            mps_info.set_bond_dimension_using_occ(
                bond_dims[0], VectorDouble(occs), bias=bias)
        mps = MultiMPS(self.n_sites, 0, 2, nroots)
        if weights is not None:
            mps.weights = VectorDouble([float(x) for x in weights])
        mps.initialize(mps_info)
        mps.random_canonicalize()

        mps.save_mutable()
        mps.deallocate()
        mps_info.save_mutable()
        mps_info.deallocate_mutable()

        # MPO
        tx = time.perf_counter()
        mpo = MPOQC(self.hamil, QCTypes.Conventional)
        mpo = SimplifiedMPO(mpo, RuleQC(), True, True,
                            OpNamesSet((OpNames.R, OpNames.RD)))
        self.mpo_orig = mpo

        if self.mpi is not None:
            mpo = ParallelMPO(mpo, self.prule)

        if self.verbose >= 3:
            _print('MPO time = ', time.perf_counter() - tx)

        if self.print_statistics:
            _print('MPO BOND DIMS = ', ''.join(
                ["%6d" % (x.m * x.n) for x in mpo.left_operator_names]))
            max_d = max(bond_dims)
            mps_info2 = MPSInfo(self.n_sites, self.hamil.vacuum,
                                self.target, self.hamil.basis)
            mps_info2.set_bond_dimension(max_d)
            _, mem2, disk = mpo.estimate_storage(mps_info2, 2)
            _print("INIT MPS BOND DIMS = ", ''.join(
                ["%6d" % x.n_states_total for x in mps_info.left_dims]))
            _print("EST MAX MPS BOND DIMS = ", ''.join(
                ["%6d" % x.n_states_total for x in mps_info2.left_dims]))
            _print("EST PEAK MEM = ", SIDMRG.fmt_size(
                mem2), " SCRATCH = ", SIDMRG.fmt_size(disk))
            mps_info2.deallocate_mutable()
            mps_info2.deallocate()

        # DMRG
        me = MovingEnvironment(mpo, mps, mps, "DMRG")
        if self.dctr:
            me.delayed_contraction = OpNamesSet.normal_ops()
        tx = time.perf_counter()
        me.delayed_contraction = OpNamesSet.normal_ops()
        me.cached_contraction = True
        me.save_partition_info = True
        me.init_environments(self.verbose >= 4)
        if self.verbose >= 3:
            _print('DMRG INIT time = ', time.perf_counter() - tx)
        dmrg = DMRG(me, VectorUBond(bond_dims), VectorDouble(noises))
        dmrg.davidson_conv_thrds = VectorDouble(dav_thrds)
        dmrg.noise_type = NoiseTypes.ReducedPerturbativeCollected
        dmrg.decomp_type = DecompositionTypes.DensityMatrix
        dmrg.iprint = max(self.verbose - 1, 0)
        dmrg.cutoff = cutoff
        dmrg.solve(n_steps, mps.center == 0, conv_tol)

        self.energies = dmrg.energies[-1]
        self.bond_dim = bond_dims[-1]

        mps.save_data()
        mps_info.save_data(self.scratch + "/MPS_INFO.%s" % tag)
        mps_info.deallocate()

        if self.print_statistics:
            dmain, dseco, imain, iseco = Global.frame.peak_used_memory
            _print("PEAK MEM USAGE:",
                  "DMEM = ", SIDMRG.fmt_size(dmain + dseco),
                  "(%.0f%%)" % (dmain * 100 / (dmain + dseco)),
                  "IMEM = ", SIDMRG.fmt_size(imain + iseco),
                  "(%.0f%%)" % (imain * 100 / (imain + iseco)))

        if self.verbose >= 1:
            _print(("=== Energy = " + ("%15.8f") * len(self.energies)) %
                  tuple(self.energies))

        if self.verbose >= 2:
            _print('>>> COMPLETE DMRG | Time = %.2f <<<' %
                  (time.perf_counter() - t))

        return self.energies

    def prepare_mps(self, tags, iroots=None):
        """Separate state-averaged MPS for later treatment.

        Args:
            tags : list(str)
                MPS tags to include.
            iroots : list(list(int)) or None
                For each tag, which root to include.
                If None, all roots are included.
        """
        if self.mpi is not None:
            self.mpi.barrier()

        impo = IdentityMPO(self.hamil)
        impo = SimplifiedMPO(impo, NoTransposeRule(RuleQC()),
                             True, True, OpNamesSet((OpNames.R, OpNames.RD)))
        if self.mpi is not None:
            impo = ParallelMPO(impo, self.prule)

        mpss = []
        assert iroots is None or len(iroots) == len(tags)
        for it, tag in enumerate(tags):
            mps_info = MultiMPSInfo(0)
            mps_info.load_data(scratch + "/MPS_INFO.%s" % tag)
            mps = MultiMPS(mps_info)
            mps.load_data()
            if iroots is None:
                troots = range(mps.nroots)
            else:
                troots = iroots[it]
            for ir in troots:
                smps = mps.extract(ir, mps_info.tag + "-%d" % ir)
                if self.mpi is not None:
                    self.mpi.barrier()
                if smps.center != 0:
                    me = MovingEnvironment(impo, smps, smps, "EX")
                    me.delayed_contraction = OpNamesSet.normal_ops()
                    me.cached_contraction = True
                    me.save_partition_info = True
                    me.init_environments(0)
                    expect = Expect(me, smps.info.bond_dim +
                                    100, smps.info.bond_dim + 100)
                    expect.iprint = 0
                    expect.solve(True, False)
                    smps.save_data()
                mpss.append(smps)

        impo.deallocate()
        return mpss

    def energy_expectation(self, mpss):
        """Energy expectation on MPS."""

        if self.mpi is not None:
            self.mpi.barrier()

        # MPO
        mpo = MPOQC(self.hamil, QCTypes.Conventional)
        mpo = SimplifiedMPO(mpo, RuleQC(), True, True,
                            OpNamesSet((OpNames.R, OpNames.RD)))
        if self.mpi is not None:
            mpo = ParallelMPO(mpo, self.prule)

        eners = []
        for mps in mpss:
            mps.load_data()
            me = MovingEnvironment(mpo, mps, mps, "EX")
            me.delayed_contraction = OpNamesSet.normal_ops()
            me.cached_contraction = True
            me.save_partition_info = True
            me.init_environments(self.verbose >= 4)
            expect = Expect(me, mps.info.bond_dim + 100,
                            mps.info.bond_dim + 100)
            expect.iprint = max(self.verbose - 1, 0)
            ener = expect.solve(False, mps.center == 0)
            mps.save_data()
            _print("TAG = %5s TARGET = %r EXPT = %15.8f" %
                  (mps.info.tag, mps.info.targets, ener))
            eners.append(ener)

        mpo.deallocate()
        return eners

    def trans_onepdm(self, mpss, soc=True, has_tran=True):
        """
        transition one-particle density matrix
        """
        if self.mpi is not None:
            self.mpi.barrier()

        # MPO
        pmpo = PDM1MPOQC(self.hamil, 1 if soc else 0)
        pmpo = SimplifiedMPO(pmpo,
                             NoTransposeRule(
                                 RuleQC()) if has_tran else RuleQC(),
                             True, True, OpNamesSet((OpNames.R, OpNames.RD)))
        if self.mpi is not None:
            pmpo = ParallelMPO(pmpo, self.pdmrule)

        pdms = np.zeros((len(mpss), len(mpss)), dtype=object)
        for sbra, bmps_orig in enumerate(mpss):
            for sket, kmps_orig in enumerate(mpss):
                bmps = bmps_orig.extract(0, 'TMP-BRA')
                kmps = kmps_orig.extract(0, 'TMP-KET')
                if self.mpi is not None:
                    self.mpi.barrier()
                assert bmps.center == 0 and kmps.center == 0
                bmps.load_data()
                kmps.load_data()
                me = MovingEnvironment(pmpo, bmps, kmps, "EX")
                me.delayed_contraction = OpNamesSet.normal_ops()
                me.cached_contraction = True
                me.save_partition_info = True
                me.init_environments(self.verbose >= 4)
                expect = Expect(me, bmps.info.bond_dim + 100,
                                kmps.info.bond_dim + 100)
                expect.iprint = max(self.verbose - 1, 0)
                expect.solve(True, kmps.center == 0)
                if SpinLabel == SU2:
                    dmr = expect.get_1pdm_spatial(self.n_sites)
                    dm = np.array(dmr).copy()
                    qsbra = bmps.info.targets[0].twos
                    # fix different Wigner–Eckart theorem convention
                    dm *= np.sqrt(qsbra + 1)
                else:
                    dmr = expect.get_1pdm(self.n_sites)
                    dm = np.array(dmr).copy()
                    dm = dm.reshape((self.n_sites, 2,
                                     self.n_sites, 2))
                    dm = np.transpose(dm, (0, 2, 1, 3))
                dmr.deallocate()
                _print("IBRA = %2d (%5s - %10r) IKET = %2d (%5s - %10r) TRACE = %15.8f" % (sbra,
                                                                                          bmps_orig.info.tag, bmps.info.targets[
                                                                                              0], sket, kmps_orig.info.tag,
                                                                                          kmps.info.targets[0], np.diag(dm).sum()))
                pdms[sbra, sket] = dm / np.sqrt(2)

        pmpo.deallocate()
        return pdms

    def onepdm(self, mpss, soc=False, has_tran=False):
        """
        one-particle density matrix
        pdm[i, j] -> <AD_{i,alpha} A_{j,alpha}> + <AD_{i,beta} A_{j,beta}>
        """

        if self.mpi is not None:
            self.mpi.barrier()

        # MPO
        pmpo = PDM1MPOQC(self.hamil, 1 if soc else 0)
        pmpo = SimplifiedMPO(pmpo,
                             NoTransposeRule(
                                 RuleQC()) if has_tran else RuleQC(),
                             True, True, OpNamesSet((OpNames.R, OpNames.RD)))
        if self.mpi is not None:
            pmpo = ParallelMPO(pmpo, self.pdmrule)

        pdms = []
        for mps_orig in mpss:
            mps = mps_orig.extract(0, 'TMP-KET')
            if self.mpi is not None:
                self.mpi.barrier()
            assert mps.center == 0
            mps.load_data()
            me = MovingEnvironment(pmpo, mps, mps, "EX")
            me.delayed_contraction = OpNamesSet.normal_ops()
            me.cached_contraction = True
            me.save_partition_info = True
            me.init_environments(self.verbose >= 4)
            expect = Expect(me, mps.info.bond_dim + 100,
                            mps.info.bond_dim + 100)
            expect.iprint = max(self.verbose - 1, 0)
            expect.solve(True, mps.center == 0)
            if SpinLabel == SU2:
                dmr = expect.get_1pdm_spatial(self.n_sites)
                dm = np.array(dmr).copy()
            else:
                dmr = expect.get_1pdm(self.n_sites)
                dm = np.array(dmr).copy()
                dm = dm.reshape((self.n_sites, 2,
                                 self.n_sites, 2))
                dm = np.transpose(dm, (0, 2, 1, 3))
            dmr.deallocate()
            _print("TAG = %5s TARGET = %r TRACE = %15.8f" % (mps.info.tag,
                                                            mps.info.targets, np.diag(dm).sum()))
            pdms.append(dm)

        pmpo.deallocate()
        return pdms

    def __del__(self):
        if self.hamil is not None:
            self.hamil.deallocate()
        if self.fcidump is not None:
            self.fcidump.deallocate()
        if self.mpo_orig is not None:
            self.mpo_orig.deallocate()
        release_memory()


def get_jk(mol, dm0):
    hso2e = mol.intor('int2e_p1vxp1', 3).reshape(
        3, mol.nao, mol.nao, mol.nao, mol.nao)
    vj = np.einsum('yijkl,lk->yij', hso2e, dm0)
    vk = np.einsum('yijkl,jk->yil', hso2e, dm0)
    vk += np.einsum('yijkl,li->ykj', hso2e, dm0)
    return vj, vk


def get_jk_amfi(mol, dm0):
    '''Atomic-mean-field approximation'''
    ao_loc = mol.ao_loc_nr()
    nao = ao_loc[-1]
    vj = np.zeros((3, nao, nao))
    vk = np.zeros((3, nao, nao))
    atom = copy.copy(mol)
    aoslice = mol.aoslice_by_atom(ao_loc)
    for ia in range(mol.natm):
        b0, b1, p0, p1 = aoslice[ia]
        atom._bas = mol._bas[b0:b1]
        vj1, vk1 = get_jk(atom, dm0[p0:p1, p0:p1])
        vj[:, p0:p1, p0:p1] = vj1
        vk[:, p0:p1, p0:p1] = vk1
    return vj, vk

# hso (complex, pure imag) in unit cm-1


def compute_hso_ao(mol, dm0, qed_fac=1, amfi=False):
    from pyscf.data import nist
    alpha2 = nist.ALPHA ** 2
    hso1e = mol.intor_asymmetric('int1e_pnucxp', 3)
    vj, vk = get_jk_amfi(mol, dm0) if amfi else get_jk(mol, dm0)
    hso2e = vj - vk * 1.5
    hso = qed_fac * (alpha2 / 4) * (hso1e + hso2e)
    return hso * 1j

# separate T^1 to T^1_(-1,0,1)


def spin_proj(cg, pdm, tjo, tjb, tjk):
    nmo = pdm.shape[0]
    ppdm = np.zeros((tjb + 1, tjk + 1, tjo + 1, nmo, nmo))
    for ibra in range(tjb + 1):
        for iket in range(tjk + 1):
            for iop in range(tjo + 1):
                tmb = -tjb + 2 * ibra
                tmk = -tjk + 2 * iket
                tmo = -tjo + 2 * iop
                factor = (-1) ** ((tjb - tmb) // 2) * \
                    cg.wigner_3j(tjb, tjo, tjk, -tmb, tmo, tmk)
                if factor != 0:
                    ppdm[ibra, iket, iop] = pdm * factor
    return ppdm

# from T^1_(-1,0,1) to Tx, Ty, Tz


def xyz_proj(ppdm):
    xpdm = np.zeros(ppdm.shape, dtype=complex)
    xpdm[:, :, 0] = (0.5 + 0j) * (ppdm[:, :, 0] - ppdm[:, :, 2])
    xpdm[:, :, 1] = (0.5j + 0) * (ppdm[:, :, 0] + ppdm[:, :, 2])
    xpdm[:, :, 2] = (np.sqrt(0.5) + 0j) * ppdm[:, :, 1]
    return xpdm


do_h2o = False
do_cu_atom = True

if __name__ == "__main__" and do_h2o:

    # parameters
    n_threads = 4
    hf_type = "RHF"  # RHF or UHF
    mpg = 'c2v'  # point group: d2h or c1
    scratch = './tmp'

    memory = 1E9  # in bytes
    verbose = 1
    do_ccsd = False  # True: use ccsd init MPS; False: use random init MPS
    bias = 10

    # if n_sweeps is larger than len(bond_dims), the last value will be repeated
    bond_dims = [250, 250, 250, 500, 500, 1000]
    # unit : norm(wfn) ** 2 (same as StackBlock)
    noises = [1E-6, 1E-7, 1E-8, 1E-9, 1E-9, 0]
    # unit : norm(wfn) ** 2 (same as StackBlock)
    dav_thrds = [1E-6, 1E-6, 1E-7, 1E-7, 1E-8, 1E-8, 1E-10]
    n_sweeps = 30
    conv_tol = 1E-14

    import os
    if MPI is None or MPI.rank == 0:
        if not os.path.isdir(scratch):
            os.makedirs(scratch)
    if MPI is not None:
        MPI.barrier()
    os.environ['TMPDIR'] = scratch

    from pyscf import gto, scf, symm, ao2mo, cc
    from pyscf.data import nist

    mol = gto.M(atom="""
        O  0.000000  0.000000  0.000000
        H  0.758602  0.000000  0.504284
        H  0.758602  0.000000  -0.504284
    """, basis='sto3g', verbose=0, symmetry=mpg)
    pg = mol.symmetry.lower()

    if hf_type == "RHF":
        mf = scf.RHF(mol)
    elif hf_type == "UHF":
        assert SpinLabel == SZ
        mf = scf.UHF(mol)

    if MPI is None or MPI.rank == 0:
        ener = mf.kernel()
    else:
        ener = 0
        mf.mo_coeff = None

    if MPI is not None:
        ener = comm.bcast(ener, root=0)
        mf.mo_coeff = comm.bcast(mf.mo_coeff, root=0)

    _print("SCF Energy = %20.15f" % ener)
    _print(("NON-" if SpinLabel == SZ else "") + "SPIN-ADAPTED")

    if pg == 'd2h':
        fcidump_sym = ["Ag", "B3u", "B2u", "B1g", "B1u", "B2g", "B3g", "Au"]
    elif pg == 'c2v':
        fcidump_sym = ['A1', 'B1', 'B2', 'A2']
    elif pg == 'c1':
        fcidump_sym = ["A"]
    else:
        raise SIDMRGError("Point group %d not supported yet!" % pg)

    mo_coeff = mf.mo_coeff

    if hf_type == "RHF":

        n_mo = mo_coeff.shape[1]

        orb_sym_str = symm.label_orb_symm(
            mol, mol.irrep_name, mol.symm_orb, mo_coeff)
        orb_sym = np.array([fcidump_sym.index(i) + 1 for i in orb_sym_str])

        h1e = mo_coeff.T @ mf.get_hcore() @ mo_coeff
        g2e = ao2mo.restore(8, ao2mo.kernel(mol, mo_coeff), n_mo)
        ecore = mol.energy_nuc()
        na = nb = mol.nelectron // 2

    else:

        mo_coeff_a, mo_coeff_b = mo_coeff[0], mo_coeff[1]
        n_mo = mo_coeff_b.shape[1]

        orb_sym_str_a = symm.label_orb_symm(
            mol, mol.irrep_name, mol.symm_orb, mo_coeff_a)
        orb_sym_a = np.array([fcidump_sym.index(i) + 1 for i in orb_sym_str_a])

        orb_sym = orb_sym_a

        h1ea = mo_coeff_a.T @ mf.get_hcore() @ mo_coeff_a
        h1eb = mo_coeff_b.T @ mf.get_hcore() @ mo_coeff_b
        g2eaa = ao2mo.restore(8, ao2mo.kernel(mol, mo_coeff_a), n_mo)
        g2ebb = ao2mo.restore(8, ao2mo.kernel(mol, mo_coeff_b), n_mo)
        g2eab = ao2mo.kernel(
            mol, [mo_coeff_a, mo_coeff_a, mo_coeff_b, mo_coeff_b])
        h1e = (h1ea, h1eb)
        g2e = (g2eaa, g2ebb, g2eab)
        ecore = mol.energy_nuc()
        na, nb = mol.nelec

    if do_ccsd:
        if MPI is None or MPI.rank == 0:
            mcc = cc.CCSD(mf)
            mcc.kernel()
            dmmo = mcc.make_rdm1()
            if hf_type == "RHF":
                occs = np.diag(dmmo)
            else:
                occs = np.diag(dmmo[0]) + np.diag(dmmo[1])
        else:
            occs = None
        if MPI is not None:
            occs = comm.bcast(occs, root=0)
        _print('OCCS = ', occs)
    else:
        occs = None

    dmrg = SIDMRG(scratch=scratch, memory=memory,
                  verbose=verbose, omp_threads=n_threads, mpi=MPI)
    dmrg.init_hamiltonian(pg, n_sites=n_mo, n_elec=na + nb, twos=na - nb, isym=1,
                          orb_sym=orb_sym, e_core=ecore, h1e=h1e, g2e=g2e,
                          save_fcidump=scratch + "/FCIDUMP")
    dmrg_opts = dict(bond_dims=bond_dims, noises=noises, dav_thrds=dav_thrds, occs=occs, bias=bias,
                     n_steps=n_sweeps, conv_tol=conv_tol)
    # SpinLabel(nelec, 2S, point group irrep)
    charge = 0
    e1 = dmrg.dmrg(target=SpinLabel(na + nb - charge, 0 +
                                    charge, 0), nroots=4, tag='MPS1', **dmrg_opts)
    e2 = dmrg.dmrg(target=SpinLabel(na + nb - charge, 2 +
                                    charge, 2), nroots=4, tag='MPS2', **dmrg_opts)
    e3 = dmrg.dmrg(target=SpinLabel(na + nb - charge, 2 +
                                    charge, 1), nroots=4, tag='MPS3', **dmrg_opts)
    e4 = dmrg.dmrg(target=SpinLabel(na + nb - charge, 2 +
                                    charge, 3), nroots=4, tag='MPS4', **dmrg_opts)
    eners = np.concatenate([e1, e2, e3, e4])
    mpss = dmrg.prepare_mps(tags=['MPS1', 'MPS2', 'MPS3', 'MPS4'])
    dmmo = dmrg.onepdm(mpss, )
    mo_coeff_inv = np.linalg.inv(mo_coeff)
    dm0ao = mo_coeff_inv.T @ dmmo[0] @ mo_coeff_inv
    hsoao = compute_hso_ao(mol, dm0ao, amfi=True) * 2
    if True:
        v = 2.1281747964476273E-004j * 2
        hsoao[:] = 0
        hsoao[0, 4, 3] = hsoao[1, 2, 4] = hsoao[2, 3, 2] = v
        hsoao[0, 3, 4] = hsoao[1, 4, 2] = hsoao[2, 2, 3] = -v
    hso = np.einsum('rij,ip,jq->rpq', hsoao, mo_coeff, mo_coeff)
    cg = CG(200)
    cg.initialize()
    print('HSO.SHAPE = ', hso.shape)
    pdms = dmrg.trans_onepdm(mpss)
    print('PDMS.SHAPE = ', pdms.shape)
    thrds = 29.0  # cm-1
    au2cm = nist.HARTREE2J / nist.PLANCK / nist.LIGHT_SPEED_SI * 1e-2
    n_mstates = sum([mpss[ibra].info.targets[0].twos +
                     1 for ibra in range(len(pdms))])
    hsiso = np.zeros((n_mstates, n_mstates), dtype=complex)
    hdiag = np.zeros((n_mstates, ), dtype=complex)
    qls = []
    imb = 0
    for ibra in range(len(pdms)):
        imk = 0
        tjb = mpss[ibra].info.targets[0].twos
        for iket in range(len(pdms)):
            pdm = pdms[ibra, iket]
            tjk = mpss[iket].info.targets[0].twos
            xpdm = xyz_proj(spin_proj(cg, pdm, 2, tjb, tjk))
            for ibm in range(xpdm.shape[0]):
                for ikm in range(xpdm.shape[1]):
                    somat = np.einsum('rij,rij->', xpdm[ibm, ikm], hso)
                    hsiso[ibm + imb, ikm + imk] = somat
                    somat *= au2cm
                    if abs(somat) > thrds:
                        print(('I1 = %4d (E1 = %15.8f) S1 = %4.1f MS1 = %4.1f '
                               + 'I2 = %4d (E2 = %15.8f) S2 = %4.1f MS2 = %4.1f Re = %9.3f Im = %9.3f')
                              % (ibra, eners[ibra], tjb / 2, -tjb / 2 + ibm, iket, eners[iket], tjk / 2,
                                 -tjk / 2 + ikm, somat.real, somat.imag))
            imk += tjk + 1
        for ibm in range(tjb + 1):
            qls.append((ibra, eners[ibra], tjb / 2, -tjb / 2 + ibm))
        hdiag[imb:imb + tjb + 1] = eners[ibra]
        imb += tjb + 1
    symm_err = np.linalg.norm(np.abs(hsiso - hsiso.T.conj()))
    print('SYMM Error (should be small) = ', symm_err)
    assert symm_err < 1E-10
    hfull = hsiso + np.diag(hdiag)
    heig, hvec = np.linalg.eigh(hfull)
    print('Total energies including SO-coupling:\n')
    for i in range(len(heig)):
        shvec = np.zeros(len(eners))
        imb = 0
        for ibra in range(len(eners)):
            tjb = mpss[ibra].info.targets[0].twos
            shvec[ibra] = np.linalg.norm(hvec[imb:imb + tjb + 1, i]) ** 2
            imb += tjb + 1
        iv = np.argmax(np.abs(shvec))
        print(' State %4d Total energy: %15.8f | largest |coeff|**2 %10.6f from I = %4d E = %15.8f S = %4.1f'
              % (i, heig[i], shvec[iv], iv, eners[iv], mpss[iv].info.targets[0].twos / 2))

    del dmrg  # IMPORTANT!!! --> to release stack memory


if __name__ == "__main__" and do_cu_atom:

    # parameters
    n_threads = 14
    mpg = 'c1'  # point group: d2h or c1
    scratch = './tmp'

    memory = 20E9  # in bytes
    verbose = 3
    bias = 5

    # if n_sweeps is larger than len(bond_dims), the last value will be repeated
    bond_dims = [250, 250, 250, 500, 500, 500, 1000, 1000, 1000]
    # unit : norm(wfn) ** 2 (same as StackBlock)
    noises = [1E-4, 1E-4, 1E-5, 1E-5, 1E-6, 1E-6, 1E-6, 1E-6, 0]
    # unit : norm(wfn) ** 2 (same as StackBlock)
    dav_thrds = [1E-5, 1E-5, 1E-6, 1E-6, 1E-7, 1E-7, 1E-7, 1E-7]
    n_sweeps = 10
    conv_tol = 1E-6

    import os
    if MPI is None or MPI.rank == 0:
        if not os.path.isdir(scratch):
            os.makedirs(scratch)
    if MPI is not None:
        MPI.barrier()
    os.environ['TMPDIR'] = scratch

    from pyscf import gto, scf, symm, ao2mo, cc, mcscf, tools
    from pyscf.mcscf import casci_symm
    from pyscf.data import nist

    ano = gto.basis.parse("""
    #BASIS SET: (21s,15p,10d,6f,4g) -> [6s,5p,3d,2f]
    Cu    S
        9148883.               0.00011722            -0.00003675             0.00001383            -0.00000271             0.00000346            -0.00001401
        1369956.               0.00033155            -0.00010403             0.00003914            -0.00000766             0.00000979            -0.00003959
        311782.6               0.00088243            -0.00027733             0.00010439            -0.00002047             0.00002626            -0.00010612
        88318.80               0.00217229            -0.00068417             0.00025745            -0.00005030             0.00006392            -0.00025900
        28815.53               0.00529547            -0.00167686             0.00063203            -0.00012442             0.00016087            -0.00064865
        10403.46               0.01296630            -0.00413460             0.00155718            -0.00030269             0.00038023            -0.00154770
        4057.791               0.03188050            -0.01035127             0.00391659            -0.00077628             0.00101716            -0.00408342
        1682.974               0.07576569            -0.02534791             0.00959476            -0.00185240             0.00229152            -0.00940206
        733.7543               0.16244637            -0.05836947             0.02238837            -0.00448270             0.00597497            -0.02389117
        333.2677               0.28435959            -0.11673284             0.04528606            -0.00866024             0.01044695            -0.04359832
        156.4338               0.34173643            -0.18466860             0.07498378            -0.01544726             0.02151583            -0.08588857
        74.69721               0.20955087            -0.14700899             0.06188170            -0.01089864             0.01030128            -0.04734865
        33.32262               0.03544751             0.18483257            -0.09047327             0.01467867            -0.00984324             0.06382592
        16.62237              -0.00243996             0.56773609            -0.39288472             0.09154122            -0.14292630             0.59363541
        8.208260               0.00146990             0.37956092            -0.35664956             0.06330857            -0.05612707             0.41506062
        3.609400              -0.00064379             0.04703755             0.34554793            -0.06093879             0.09135650            -1.70027942
        1.683449               0.00024738            -0.00072783             0.70639197            -0.27148545             0.45831402            -0.29799956
        0.733757              -0.00008118             0.00102976             0.25335911            -0.10138944            -0.19726740             1.70719425
        0.110207               0.00001615            -0.00007020             0.00658749             0.71212180            -1.30310157            -0.65600053
        0.038786              -0.00001233             0.00007090            -0.00044247             0.34613175             0.87975568            -0.31924524
        0.015514               0.00000452            -0.00002454             0.00072338             0.07198709             0.57310889             0.61772456
    Cu    P
    9713.253                   0.00039987            -0.00014879             0.00005053            -0.00015878             0.00012679
    2300.889                   0.00210157            -0.00078262             0.00026719            -0.00085359             0.00068397
        746.7706               0.01001097            -0.00376229             0.00127718            -0.00399919             0.00320291
        284.6806               0.03836039            -0.01457968             0.00499760            -0.01608092             0.01301351
        119.9999               0.11626756            -0.04571093             0.01561376            -0.04924864             0.04021371
        54.07386               0.25899831            -0.10593013             0.03689701            -0.12165679             0.10150109
        25.37321               0.38428226            -0.16836228             0.05796469            -0.17713367             0.13700954
        12.20962               0.29911210            -0.10373213             0.03791966            -0.14448197             0.10628972
        5.757421               0.08000672             0.21083155            -0.11706499             0.64743311            -0.83390265
        2.673402               0.00319440             0.48916443            -0.20912257             0.54797692             0.14412002
        1.186835               0.00147252             0.38468638            -0.05800532            -0.78180321             0.98301239
        0.481593              -0.00023391             0.08529338             0.16430736            -0.52607778            -0.52956418
        0.192637               0.00020761            -0.00161045             0.52885466             0.22437177            -0.75303984
        0.077055              -0.00008963             0.00206530             0.37373016             0.33392453             0.44288700
        0.030822               0.00002783            -0.00064175             0.08399866             0.14446374             0.57511589
    Cu    D
        249.3497               0.00134561            -0.00158359             0.00191116
        74.63837               0.01080983            -0.01281116             0.01772495
        28.37641               0.04733793            -0.05642679             0.07056381
        11.94893               0.13772582            -0.16641393             0.22906101
        5.317646               0.26263833            -0.35106014             0.43938511
        2.364417               0.33470401            -0.23717890            -0.26739908
        1.012386               0.31000597             0.21994641            -0.62590624
        0.406773               0.21316819             0.48841363             0.12211340
        0.147331               0.07865365             0.29537131             0.53683629
        0.058932               0.00603096             0.04295539             0.19750055
    Cu    F
        15.4333                0.03682063            -0.06200336
        6.1172                 0.24591418            -0.52876669
        2.4246                 0.50112688            -0.29078386
        0.9610                 0.35850648             0.52124157
        0.3809                 0.14728062             0.37532205
        0.1510                 0.02564748             0.10659798
    """)

    mol = gto.M(atom='Cu 0 0 0',
                symmetry=mpg,
                basis=ano, spin=1, charge=0, verbose=5)
    mf = scf.newton(scf.RHF(mol).sfx2c1e())
    pg = mol.symmetry.lower()

    if MPI is None or MPI.rank == 0:
        ener = mf.kernel()
    else:
        ener = 0
        mf.mo_coeff = None

    if MPI is not None:
        ener = comm.bcast(ener, root=0)
        mf.mo_coeff = comm.bcast(mf.mo_coeff, root=0)

    _print("SCF Energy = %20.15f" % ener)
    _print(("NON-" if SpinLabel == SZ else "") + "SPIN-ADAPTED")

    cas_list = [9, 10, 11, 12, 13, 14, 22, 23, 24, 25, 26]
    mc = mcscf.CASSCF(mf, 11, (6, 5))
    mo = mcscf.sort_mo(mc, mf.mo_coeff, cas_list, base=0)
    mc.state_average_((0.5, 0.1, 0.1, 0.1, 0.1, 0.1))
    mc.fix_spin_(ss=0.75)

    if MPI is None or MPI.rank == 0:
        mc.kernel(mo)
        eners = mc.e_states
        dmao = mc.make_rdm1()
    else:
        mc.mo_coeff = None
        eners = None
        dmao = None

    if MPI is not None:
        mc.mo_coeff = comm.bcast(mc.mo_coeff, root=0)
        eners = comm.bcast(eners, root=0)
        dmao = comm.bcast(dmao, root=0)

    _print("CASSCF Energies =", "".join(["%20.10f" % x for x in eners]))

    # wfn (nroots, multiplicity, irrep)
    states = [(6, 2, 'A')]
    weights = [(0.5, ) + (0.1, ) * 5]
    tags = ['MPS%d' % x for x in range(len(states))]
    # active space (n_orbirals, n_electrons)
    # cas = (45, 19)
    cas = (11, 11)
    nactorb, nelec = cas
    mf.mo_coeff = mc.mo_coeff
    mo_coeff = mf.mo_coeff
    ff = mcscf.CASCI(mf, cas[0], cas[1])
    ncore = ff.ncore
    ff.mo_coeff = casci_symm.label_symmetry_(ff, mf.mo_coeff)
    orbsym = ff.mo_coeff.orbsym
    orbsym = [tools.fcidump.ORBSYM_MAP[mol.groupname][i]
              for i in orbsym[ncore:ncore + nactorb]]
    _print(ff.mo_coeff.orbsym[ncore:ncore + nactorb])
    h1e, e_core = ff.get_h1cas()
    g2e = ao2mo.restore(1, ff.get_h2cas(), nactorb)
    if MPI is not None:
        ff.mo_coeff = comm.bcast(ff.mo_coeff, root=0)
        e_core = comm.bcast(e_core, root=0)
        h1e = comm.bcast(h1e, root=0)
        g2e = comm.bcast(g2e, root=0)

    _print('Integrals = ', h1e.shape, g2e.shape, e_core)

    dmrg = SIDMRG(scratch=scratch, memory=memory,
                  verbose=verbose, omp_threads=n_threads, mpi=MPI)
    dmrg.init_hamiltonian(pg, n_sites=nactorb, n_elec=nelec, twos=mol.spin, isym=1,
                          orb_sym=orbsym, e_core=e_core, h1e=h1e, g2e=g2e,
                          save_fcidump=scratch + "/FCIDUMP")
    dmrg_opts = dict(bond_dims=bond_dims, noises=noises, dav_thrds=dav_thrds,
                     occs=None, bias=bias, n_steps=n_sweeps, conv_tol=conv_tol)

    eners = [None] * len(states)
    for istate, state in enumerate(states):
        nroots = state[0]
        mult = state[1]
        wfnsym = state[2]
        tag = tags[istate]
        ms2 = mult - 1
        na = (nelec + ms2) // 2
        nb = nelec - na
        ff.nelecas = (na, nb)
        isym = mol.irrep_name.index(wfnsym)
        eners[istate] = dmrg.dmrg(target=SpinLabel(na + nb, na - nb, isym),
                                  nroots=nroots, tag=tag, weights=weights[istate], **dmrg_opts)

    eners = np.concatenate(eners)
    mpss = dmrg.prepare_mps(tags=tags)
    mo_coeff_inv = np.linalg.inv(mo_coeff)
    _print('\nGenerating Spin-Orbit Integrals:\n')
    if MPI is None or MPI.rank == 0:
        hsoao = compute_hso_ao(mol, dmao, amfi=True) * 2
        hso = np.einsum('rij,ip,jq->rpq', hsoao,
                        mo_coeff[:, ncore:ncore + nactorb],
                        mo_coeff[:, ncore:ncore + nactorb])
    else:
        hso = None
    if MPI is not None:
        hso = comm.bcast(hso, root=0)
    cg = CG(200)
    cg.initialize()
    _print('HSO.SHAPE = ', hso.shape)
    dmrg.verbose = 1
    pdms = dmrg.trans_onepdm(mpss)
    _print('PDMS.SHAPE = ', pdms.shape)
    thrds = 29.0  # cm-1
    au2cm = nist.HARTREE2J / nist.PLANCK / nist.LIGHT_SPEED_SI * 1e-2
    _print("\nComplex SO-Hamiltonian matrix elements over spin components of spin-free eigenstates:")
    _print("(In cm-1. Print threshold:%10.3f cm-1)\n" % thrds)
    n_mstates = sum([mpss[ibra].info.targets[0].twos +
                     1 for ibra in range(len(pdms))])
    hsiso = np.zeros((n_mstates, n_mstates), dtype=complex)
    hdiag = np.zeros((n_mstates, ), dtype=complex)
    qls = []
    imb = 0
    for ibra in range(len(pdms)):
        imk = 0
        tjb = mpss[ibra].info.targets[0].twos
        for iket in range(len(pdms)):
            pdm = pdms[ibra, iket]
            tjk = mpss[iket].info.targets[0].twos
            xpdm = xyz_proj(spin_proj(cg, pdm, 2, tjb, tjk))
            for ibm in range(xpdm.shape[0]):
                for ikm in range(xpdm.shape[1]):
                    somat = np.einsum('rij,rij->', xpdm[ibm, ikm], hso)
                    hsiso[ibm + imb, ikm + imk] = somat
                    somat *= au2cm
                    if abs(somat) > thrds:
                        _print(('I1 = %4d (E1 = %15.8f) S1 = %4.1f MS1 = %4.1f '
                               + 'I2 = %4d (E2 = %15.8f) S2 = %4.1f MS2 = %4.1f Re = %9.3f Im = %9.3f')
                              % (ibra, eners[ibra], tjb / 2, -tjb / 2 + ibm, iket, eners[iket], tjk / 2,
                                 -tjk / 2 + ikm, somat.real, somat.imag))
            imk += tjk + 1
        for ibm in range(tjb + 1):
            qls.append((ibra, eners[ibra], tjb / 2, -tjb / 2 + ibm))
        hdiag[imb:imb + tjb + 1] = eners[ibra]
        imb += tjb + 1
    symm_err = np.linalg.norm(np.abs(hsiso - hsiso.T.conj()))
    _print('SYMM Error (should be small) = ', symm_err)
    assert symm_err < 1E-10
    hfull = hsiso + np.diag(hdiag)
    heig, hvec = np.linalg.eigh(hfull)
    _print('\nTotal energies including SO-coupling:\n')
    for i in range(len(heig)):
        shvec = np.zeros(len(eners))
        imb = 0
        for ibra in range(len(eners)):
            tjb = mpss[ibra].info.targets[0].twos
            shvec[ibra] = np.linalg.norm(hvec[imb:imb + tjb + 1, i]) ** 2
            imb += tjb + 1
        iv = np.argmax(np.abs(shvec))
        _print(' State %4d Total energy: %15.8f | largest |coeff|**2 %10.6f from I = %4d E = %15.8f S = %4.1f'
              % (i, heig[i], shvec[iv], iv, eners[iv], mpss[iv].info.targets[0].twos / 2))

    e0 = np.average(heig[0:2])
    e1 = np.average(heig[2:8])
    e2 = np.average(heig[8:12])

    au2ev = 27.21139
    _print("")
    _print("E 2D(5/2)         = %10.4f eV" % ((e1 - e0) * au2ev))
    _print("E 2D(3/2)         = %10.4f eV" % ((e2 - e0) * au2ev))
    _print("2D(5/2) - 2D(3/2) = %10.4f eV" % ((e2 - e1) * au2ev))

    del dmrg  # IMPORTANT!!! --> to release stack memory




#  block2: Efficient MPO implementation of quantum chemistry DMRG
#  Copyright (C) 2021 Seunghoon Lee <seunghoonlee89@gmail.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#

"""
stochastic perturbative DMRG

Author: Seunghoon Lee, 2021
Revised: Huanchen Zhai
    Aug 13, 2021 (added openmp)
    Aug 20, 2021 (added su2)
"""

from block2 import VectorUInt8
import time
import numpy as np

# MPI
from mpi4py import MPI as MPI
comm = MPI.COMM_WORLD
mrank = MPI.COMM_WORLD.Get_rank()
msize = MPI.COMM_WORLD.Get_size()

def _print(*args, **kwargs):
    if mrank == 0:
        print(*args, **kwargs)

class SPDMRGError(Exception):
    pass

class SPDMRG:
    """
    stochastic perturbative DMRG for molecules.
    """

    def __init__(self, su2, scratch='./nodex', fcidump=None, mps_tags=[], verbose=0, use_threading=True, n_steps=20):
        """
        Memory is in bytes.
        verbose = 0 (quiet), 2 (per sweep), 3 (per iteration)
        """
        self.fcidump = None 
        self.hamil = None
        self.verbose = verbose
        self.scratch = scratch
        self.mpo_orig = None
        self.use_threading = use_threading
        self.n_steps = n_steps
        self.bdims = [0, 0]
        self.su2 = su2

        if self.su2:
            from block2.su2 import MPSInfo, MPS, UnfusedMPS, StochasticPDMRG
        else:
            from block2.sz import MPSInfo, MPS, UnfusedMPS, StochasticPDMRG

        self.Edmrg = np.float64(np.load(self.scratch + '/E_dmrg.npy'))

        if len(mps_tags) == 1 and mps_tags[0] == "KET":
            mps_tags=['ZKET', 'ZBRA']
        else:
            assert len(mps_tags) == 2

        mps1_info = MPSInfo(0)
        mps1_info.load_data(self.scratch + '/%s-mps_info.bin'%(mps_tags[0]))
        mps1 = MPS(mps1_info)
        mps1_info.load_mutable()
        max_bdim_l = max([x.n_states_total for x in mps1_info.left_dims])
        max_bdim_r = max([x.n_states_total for x in mps1_info.right_dims])
        self.bdims[0] = max(max_bdim_l, max_bdim_r)
        comm.barrier()
        if mrank == 0:
            self.change_mps_center(mps1, 0)
        comm.barrier()

        mps1.load_data()
        self.mps_psi0 = UnfusedMPS(mps1)

        mps2_info = MPSInfo(0)
        mps2_info.load_data(self.scratch + '/%s-mps_info.bin'%(mps_tags[1]))
        mps2 = MPS(mps2_info)
        mps2_info.load_mutable()
        max_bdim_l = max([x.n_states_total for x in mps2_info.left_dims])
        max_bdim_r = max([x.n_states_total for x in mps2_info.right_dims])
        self.bdims[1] = max(max_bdim_l, max_bdim_r)
        comm.barrier()
        if mrank == 0:
            self.change_mps_center(mps2, 0)
        comm.barrier()
        mps2.load_data()
        self.mps_qvpsi0 = UnfusedMPS(mps2)

        self.norm_qvpsi0  = np.float64(np.load(self.scratch + '/cps_overlap.npy'))
        self.norm_qvpsi0  = self.norm_qvpsi0*self.norm_qvpsi0

        self.SPDMRG = StochasticPDMRG(self.mps_psi0, self.mps_qvpsi0, self.norm_qvpsi0) 
        self.n_sites = self.SPDMRG.n_sites
        if fcidump is not None:
            self.fcidump = fcidump
            E_dmrg = np.float64(np.load(scratch + '/E_dmrg.npy'))
            E_cas  = E_dmrg - self.fcidump.const_e 
            dm_e_pqqp = np.load(scratch + '/e_pqqp.npy')
            dm_e_pqpq = np.load(scratch + '/e_pqpq.npy')
            one_pdm = np.load(scratch + "/1pdm.npy")

            E_0 = self.SPDMRG.energy_zeroth(fcidump, dm_e_pqqp, dm_e_pqpq, one_pdm[0]+one_pdm[1])
            self.fcidump.const_e = - 0.5 * ( E_cas + E_0 ) 

    def change_mps_center(self, ket, center):
        if self.su2:
            from block2.su2 import CG
        else:
            from block2.sz import CG
        ket.load_data()
        cf = ket.canonical_form
        _print('CF = %s CENTER = %d' % (cf, ket.center))
        cg = CG(200)
        cg.initialize()
        if ket.center == center:
            if ket.canonical_form[0] == 'S':
                ket.move_right(cg, None)
                ket.move_left(cg, None)
        elif center == 0:
            if ket.center == ket.n_sites - 2:
                ket.center += 1
            if ket.canonical_form[-1] == 'C':
                ket.canonical_form = ket.canonical_form[:-1] + 'S'
            while ket.center != 0:
                ket.move_left(cg, None)
        else:
            if ket.canonical_form[0] == 'C':
                ket.canonical_form = 'K' + ket.canonical_form[1:]
            while ket.center != ket.n_sites - 1:
                ket.move_right(cg, None)
            ket.center -= 1
        if self.verbose >= 2:
            _print('--> %s' % (ket.canonical_form))
        ket.save_data()
    
    def compute_correction(self, max_samp_c, max_samp, H00, H00_2, H11, H11_2, H10, H10_2):
        max_samp_c = np.float64(max_samp_c)
        max_samp = np.float64(max_samp)
        avg_Cterm = H00
        std_Cterm = np.sqrt(abs(H00_2 - avg_Cterm * avg_Cterm) / max_samp_c)
        avg_Aterm = H11
        std_Aterm = np.sqrt(abs(H11_2 - avg_Aterm * avg_Aterm) / max_samp)
        avg_Bterm = H10
        std_Bterm = np.sqrt(abs(H10_2 - avg_Bterm * avg_Bterm) / max_samp)
        Emp2 = - avg_Aterm + avg_Bterm ** 2 / avg_Cterm
        with np.errstate(divide='ignore', invalid='ignore'):
            if abs(avg_Bterm) > 1e-10:
                std_Emp2 = std_Aterm + avg_Bterm ** 2 / abs(avg_Cterm) \
                            * (2 * std_Bterm / abs(avg_Bterm) + std_Cterm / abs(avg_Cterm) ) 
            else:
                std_Emp2 = std_Aterm
        if self.verbose >= 4:
            _print(" A   term = %15.10f (%15.10f)" % (avg_Aterm, std_Aterm))
            _print(" B   term = %15.10f (%15.10f)" % (avg_Bterm, std_Bterm))
            _print(" C   term = %15.10f (%15.10f)" % (avg_Cterm, std_Cterm))
        return Emp2, std_Emp2

    def kernel(self, max_samp):
        comm.barrier()
# 1] Importance Sampling of Determinant
# 1-1] C term
        if self.verbose >= 4:
            _print("1] IMPORTANT SAMPLING")
            _print("1-1] C term")
            _print("     sampling D_p with P_p = |<Phi_0|D_p>|^2")
            _print("     & computing <1/(E_d-E_0)>")
        Cterm = []
        H00   = 0.0
        H00_2 = 0.0
        H11   = 0.0
        H11_2 = 0.0
        H10   = 0.0
        H10_2 = 0.0
        tx = time.perf_counter()

        max_samp_per_rank = max_samp // msize
        det_string = VectorUInt8([0] * self.n_sites * 2)
        if msize == mrank - 1:
            n_samp_per_rank = max_samp - mrank * max_samp_per_rank
        else:
            n_samp_per_rank = max_samp_per_rank
        if self.use_threading:
            H00, H00_2 = self.SPDMRG.parallel_sampling(n_samp_per_rank, 0, self.fcidump)
        else:
            for num_samp in range(n_samp_per_rank):
                # sample | D_p > with P_p = |< Psi_0 | D_p >|^2
                self.SPDMRG.sampling(0, det_string)
                # calculate dE_p = < D_p | H_d | D_p > - E0
                dE_p = self.fcidump.det_energy(det_string, 0, self.n_sites) + self.fcidump.const_e
                # sampling 1/(E_p-E_0)
                H00   += 1.0 / (dE_p*np.float64(n_samp_per_rank))
                H00_2 += 1.0 / (dE_p*dE_p*np.float64(n_samp_per_rank))

        if msize != 0:
            comm.barrier()
            H00, H00_2 = [x * np.float64(n_samp_per_rank) for x in [H00, H00_2]]
            H00   = comm.allreduce(H00,   op=MPI.SUM)
            H00_2 = comm.allreduce(H00_2, op=MPI.SUM)
            H00, H00_2 = [x / np.float64(max_samp) for x in [H00, H00_2]]

        if mrank == 0 and self.verbose >= 4:
            _print("1-2] A & B term")
            _print("     sampling D_p with P_p = |<Psi_0| VQ |D_p>|^2")
            _print("     & computing <1/(E_d-E_0)> and <<D_p|Psi_0> / {(E_d-E_0) <D_p|QV|Psi_0>}>")

        Aterm = []
        Bterm = []
        if self.use_threading:
            _print("Sampling | BRA bond dimension = %d | KET bond dimension = %d | Nsample = %d"
                % (self.bdims[1], self.bdims[0], max_samp), flush=True)
            n_steps = 10
            sub_results = [0, 0, 0, 0]
            current_max_samp = 0
            tg = time.perf_counter()
            for i_samp in range(self.n_steps):
                if i_samp == self.n_steps - 1:
                    n_sub_samp = n_samp_per_rank - i_samp * (n_samp_per_rank // self.n_steps)
                else:
                    n_sub_samp = n_samp_per_rank // self.n_steps
                tx = time.perf_counter()
                results = self.SPDMRG.parallel_sampling(n_sub_samp, 1, self.fcidump)
                sub_results = [ra + rb * n_sub_samp for ra, rb in zip(sub_results,results)]
                current_max_samp += n_sub_samp
                fcms = np.float64(current_max_samp)
                part_f, part_err = self.compute_correction(max_samp, current_max_samp,
                    H00, H00_2, sub_results[0] / fcms, sub_results[1] / fcms,
                    sub_results[2] / fcms, sub_results[3] / fcms)
                print('Rank = %3d / %3d Step = %3d / %3d .. Nsample = %10d F = %18.10f Error = %9.2E T = %.2f' %
                    (mrank, msize, i_samp, self.n_steps, n_sub_samp, part_f, part_err, time.perf_counter() - tx), flush=True)
            _print("Time elapsed = %.3f" % (time.perf_counter() - tg))

            H11, H11_2, H10, H10_2 = [rx / n_samp_per_rank for rx in sub_results]
        else:
            if max_samp_per_rank > 10:
                print_samp = max_samp_per_rank // 10 
            else:
                print_samp = max_samp_per_rank  
            for num_samp in range(n_samp_per_rank):
                if num_samp % print_samp ==0:
                    print( 'processor %2d: sampling %2d %% done'%(mrank, (num_samp//print_samp+0)*10) )
                # sample | D_p > with P_p = |< Psi_0 | VQ | D_p >|^2
                Sqv_p = self.SPDMRG.sampling(1, det_string)
                # calculate dE_p = < D_p | H_d | D_p > - E0
                dE_p = self.fcidump.det_energy(det_string, 0, self.n_sites) + self.fcidump.const_e
                # sampling 1/(E_p-E_0)
                H11   += self.norm_qvpsi0 / (dE_p*np.float64(max_samp_per_rank))
                H11_2 += self.norm_qvpsi0*self.norm_qvpsi0 / (dE_p*dE_p*np.float64(max_samp_per_rank))

                S_p    = self.SPDMRG.overlap(1, det_string)
                tmp    = self.norm_qvpsi0*S_p / (Sqv_p*dE_p)
                H10   += tmp / np.float64(max_samp_per_rank)
                H10_2 += tmp*tmp / np.float64(max_samp_per_rank)

        if msize != 0:
            comm.barrier()
            H11, H11_2, H10, H10_2 = [x * np.float64(n_samp_per_rank) for x in [H11, H11_2, H10, H10_2]]
            H11   = comm.allreduce(H11,   op=MPI.SUM)
            H11_2 = comm.allreduce(H11_2, op=MPI.SUM)
            H10   = comm.allreduce(H10,   op=MPI.SUM)
            H10_2 = comm.allreduce(H10_2, op=MPI.SUM)
            H11, H11_2, H10, H10_2 = [x / np.float64(max_samp) for x in [H11, H11_2, H10, H10_2]]

        if mrank == 0: 
            Emp2, std_Emp2 = self.compute_correction(max_samp, max_samp, H00, H00_2, H11, H11_2, H10, H10_2)
        else: 
            Emp2, std_Emp2 = 0.0, 0.0 

        if mrank == 0 and self.verbose >= 4: 
            _print("")
            _print("         ===============")
            _print("         === SUMMARY ===")
            _print("         ===============")
            _print("")
            _print("    DMRG Energy = %15.10f"%(self.Edmrg))
            _print("     MP2 Energy = %15.10f (%15.10f)"%(Emp2, std_Emp2))
            _print(" --------------------------------")
            _print(" sp-DMRG Energy = %15.10f (%15.10f)"%(self.Edmrg+Emp2, std_Emp2))

        return [Emp2, std_Emp2] 
